# This file is a part of the AnyBlok project
#
#    Copyright (C) 2016 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2017 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2019 Jean-Sebastien SUZANNE <js.suzanne@gmail.com>
#    Copyright (C) 2020 Jean-Sebastien SUZANNE <js.suzanne@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from base64 import b64encode, b64decode
from .field import Field, FieldException
from .mapper import ModelAttributeAdapter, ModelAttribute
from sqlalchemy.schema import Sequence as SA_Sequence, Column as SA_Column
from sqlalchemy import types, CheckConstraint
from sqlalchemy_utils.types.color import ColorType
from sqlalchemy_utils.types.encrypted.encrypted_type import StringEncryptedType
from sqlalchemy_utils.types.password import PasswordType, Password as SAU_PWD
from sqlalchemy_utils.types.uuid import UUIDType
from sqlalchemy_utils.types.url import URLType
from sqlalchemy_utils.types.phone_number import PhoneNumberType
from sqlalchemy_utils.types.email import EmailType
from sqlalchemy_utils.types.scalar_coercible import ScalarCoercible
from sqlalchemy_utils import JSONType
from datetime import datetime, date, timedelta
from dateutil.parser import parse
from inspect import ismethod
from anyblok.config import Configuration
from .common import sgdb_in
from json import dumps, loads
import time
import pytz
import decimal
from logging import getLogger
from hashlib import md5


pycountry = None
python_pycountry_type = None
try:
    import pycountry
    if not pycountry.countries._is_loaded:
        pycountry.countries._load()

    python_pycountry_type = pycountry.countries.data_class
except ImportError:
    pass


logger = getLogger(__name__)


def wrap_default(registry, namespace, default_val):
    """Return default wrapper

    :param registry: the current registry
    :param namespace: the namespace of the model
    :param default_val:
    :return: default wrapper
    """
    def wrapper():
        """Return wrapper

        :return: default val
        """
        Model = registry.get(namespace)
        if hasattr(Model, default_val):
            func = getattr(Model, default_val)
            if ismethod(func):
                if default_val not in Model.loaded_columns:
                    if default_val not in Model.loaded_fields:
                        return func()
                    else:
                        logger.warning(
                            "The attribute %r is already declared as a default "
                            "value on the Model %r, a field with the same name "
                            "already exists" % (default_val, namespace))
                else:
                    logger.warning(
                        "The attribute %r is already declared as a default "
                        "value on the Model %r, a column with the same name "
                        "already exists" % (default_val, namespace))
            else:
                logger.warning(
                    "The attribute %r is already declared as a default "
                    "value on the Model %r, an instance method with the same "
                    "name already exists" % (default_val, namespace))

        return default_val

    return wrapper


class ColumnDefaultValue:

    def __init__(self, callable):
        self.callable = callable

    def get_default_callable(self, registry, namespace, fieldname, properties):
        """Get default callable

        :param registry: the current registry
        :param namespace: the namespace of the model
        :param fieldname: the fieldname of the model
        :param properties: the properties of the model
        :return: default callable
        """
        return self.callable(registry, namespace, fieldname, properties)


class CompareType:

    comparators = []

    @classmethod
    def default_comparator(cls, col1, type1, col2, type2):
        if type1.__class__ is not type2.__class__:
            raise FieldException(
                "You can't add a foreign key using columns with different "
                "types {model1!s}.{col1!s}` pointing to `{model2!s}.{col2!s}` "
                "have different types  {type1!r} -> {type2!r}".format(
                    model1=col1.model_name,
                    col1=col1.attribute_name,
                    model2=col2.model_name,
                    col2=col2.attribute_name,
                    type1=type1.__class__,
                    type2=type2.__class__
                )
            )

    @classmethod
    def add_comparator(cls, type1, type2):

        def wrapper(funct):
            cls.comparators.append((type1, type2, funct))
            return funct

        return wrapper

    @classmethod
    def validate(cls, col1, type1, col2, type2):
        for (cls1, cls2, funct) in cls.comparators:
            if type1.__class__ is cls1 and type2.__class__ is cls2:
                funct(col1, type1, col2, type2)
                return

        cls.default_comparator(col1, type1, col2, type2)


class NoDefaultValue:
    pass


class Column(Field):
    """Column class

    This class can't be instantiated
    """

    use_hybrid_property = True
    foreign_key = None
    sqlalchemy_type = None
    type = None

    def __init__(self, *args, **kwargs):
        """Initialize the column

        :param label: label of this field
        :type label: str
        """
        self.forbid_instance(Column)
        assert self.sqlalchemy_type
        self.sequence = None

        if 'type_' in kwargs:
            del kwargs['type_']

        if 'foreign_key' in kwargs:
            self.foreign_key = ModelAttributeAdapter(kwargs.pop('foreign_key'))

        if 'sequence' in kwargs:
            self.sequence = SA_Sequence(kwargs.pop('sequence'))

        self.db_column_name = None
        if 'db_column_name' in kwargs:
            self.db_column_name = kwargs.pop('db_column_name')

        self.default_val = NoDefaultValue
        if 'default' in kwargs:
            self.default_val = kwargs.pop('default')

        self.encrypt_key = kwargs.pop('encrypt_key', None)
        super(Column, self).__init__(*args, **kwargs)

    def autodoc_get_properties(self):
        """Return properties list for autodoc

        :return: autodoc properties
        """
        res = super(Column, self).autodoc_get_properties()
        res['foreign_key'] = self.foreign_key
        res['DB column name'] = self.db_column_name
        res['default'] = self.default_val
        res['is crypted'] = True if self.encrypt_key else False
        return res

    autodoc_omit_property_values = Field.autodoc_omit_property_values.union((
        ('foreign_key', None),
        ('DB column name', None),
        ('default', None),
        ('is crypted', False),
    ))

    def native_type(self, registry):
        """Return the native SqlAlchemy type

        :param registry:
        :rtype: sqlalchemy native type
        """
        return self.sqlalchemy_type

    def format_foreign_key(self, registry, namespace, fieldname, args, kwargs):
        """Format a foreign key

        :param registry: the current registry
        :param args:
        :param kwargs:
        :return:
        """
        if self.foreign_key:
            CompareType.validate(
                ModelAttribute(namespace, fieldname), self,
                self.foreign_key, self.foreign_key.get_type(registry)
            )
            args = args + (self.foreign_key.get_fk(registry),)
            kwargs['info'].update({
                'foreign_key': self.foreign_key.get_fk_name(registry),
                'remote_model': self.foreign_key.model_name,
            })

        return args

    def get_sqlalchemy_mapping(self, registry, namespace, fieldname,
                               properties):
        """Return the instance of the real field

        :param registry: current registry
        :param namespace: name of the model
        :param fieldname: name of the field
        :param properties: known properties of the model
        :rtype: sqlalchemy column instance
        """
        self.format_label(fieldname)
        args = self.args
        kwargs = self.kwargs.copy()
        if 'info' not in kwargs:
            kwargs['info'] = {}
        args = self.format_foreign_key(
            registry, namespace, fieldname, args, kwargs)

        kwargs['info']['label'] = self.label
        if self.sequence:
            args = (self.sequence,) + args

        if self.db_column_name:
            db_column_name = self.db_column_name
        else:
            db_column_name = fieldname

        if self.default_val is not NoDefaultValue:
            if isinstance(self.default_val, str):
                kwargs['default'] = wrap_default(registry, namespace,
                                                 self.default_val)
            elif isinstance(self.default_val, ColumnDefaultValue):
                kwargs['default'] = self.default_val.get_default_callable(
                    registry, namespace, fieldname, properties)
            else:
                kwargs['default'] = self.default_val

        sqlalchemy_type = self.native_type(registry)

        if self.encrypt_key:
            encrypt_key = self.format_encrypt_key(registry, namespace)
            sqlalchemy_type = self.get_encrypt_key_type(
                registry, sqlalchemy_type, encrypt_key)

        return SA_Column(db_column_name, sqlalchemy_type, *args, **kwargs)

    def get_encrypt_key_type(self, registry, sqlalchemy_type, encrypt_key):
        sqlalchemy_type = StringEncryptedType(sqlalchemy_type, encrypt_key)
        if sgdb_in(registry.engine, ['MySQL', 'MariaDB']):
            sqlalchemy_type.impl = types.String(64)

        return sqlalchemy_type

    def format_encrypt_key(self, registry, namespace):
        """Format and return the encyption key

        :param registry: the current registry
        :param namespace: the namespace of the model
        :return: encrypt key
        """
        encrypt_key = self.encrypt_key
        if encrypt_key is True:
            encrypt_key = Configuration.get('default_encrypt_key')

        if not encrypt_key:
            raise FieldException(  # pragma: no cover
                "No encrypt_key defined in the configuration")

        def wrapper():
            """Return encrypt_key wrapper

            :return:
            """
            Model = registry.get(namespace)
            if hasattr(Model, encrypt_key):
                func = getattr(Model, encrypt_key)
                if ismethod(func):
                    if encrypt_key not in Model.loaded_columns:
                        if encrypt_key not in Model.loaded_fields:
                            return func()

            return encrypt_key

        return wrapper

    def must_be_declared_as_attr(self):
        """Return True if the column have a foreign key to a remote column """
        if self.foreign_key is not None:
            return True

        return False


class ForbiddenPrimaryKey:
    """Mixin to forbid primary key on column type
    """

    def get_sqlalchemy_mapping(
        self, registry, namespace, fieldname, properties
    ):
        if self.kwargs.get("primary_key") is True:
            raise FieldException(
                f"{self.__class__} column `{namespace}.{fieldname}` "
                "are not allowed as primary key"
            )
        return super().get_sqlalchemy_mapping(
            registry, namespace, fieldname, properties
        )


class Integer(Column):
    """Integer column

    ::

        from anyblok.declarations import Declarations
        from anyblok.column import Integer


        @Declarations.register(Declarations.Model)
        class Test:

            x = Integer(default=1)

    """

    def __init__(self, *args, **kwargs):
        super(Integer, self).__init__(*args, **kwargs)
        if self.kwargs.get('primary_key') is True:
            if 'autoincrement' not in self.kwargs:
                self.kwargs['autoincrement'] = True

    sqlalchemy_type = types.Integer


class BigInteger(Column):
    """Big integer column

    ::

        from anyblok.declarations import Declarations
        from anyblok.column import BigInteger


        @Declarations.register(Declarations.Model)
        class Test:

            x = BigInteger(default=1)

    """
    sqlalchemy_type = types.BigInteger


class Boolean(Column):
    """Boolean column

    ::

        from anyblok.declarations import Declarations
        from anyblok.column import Boolean


        @Declarations.register(Declarations.Model)
        class Test:

            x = Boolean(default=True)

    """
    sqlalchemy_type = types.Boolean


class Float(ForbiddenPrimaryKey, Column):
    """Float column

    ::

        from anyblok.declarations import Declarations
        from anyblok.column import Float


        @Declarations.register(Declarations.Model)
        class Test:

            x = Float(default=1.0)

    """
    sqlalchemy_type = types.Float


"""
    Added *process_result_value* at the class *DECIMAL*, because
    this method is necessary for encrypt the column
"""
types.DECIMAL.process_result_value = lambda self, value, dialect: value


class Decimal(ForbiddenPrimaryKey, Column):
    """Decimal column

    ::

        from decimal import Decimal as D
        from anyblok.declarations import Declarations
        from anyblok.column import Decimal


        @Declarations.register(Declarations.Model)
        class Test:

            x = Decimal(default=D('1.1'))

    """
    sqlalchemy_type = types.DECIMAL

    def setter_format_value(self, value):
        """Format the given value to decimal if needed

        :param value:
        :return: decimal value
        """
        if value is not None:
            if self.encrypt_key:
                value = str(value)
            elif not isinstance(value, decimal.Decimal):
                value = decimal.Decimal(value)

        return value

    def getter_format_value(self, value):
        if value is None:
            return None  # pragma: no cover

        if self.encrypt_key:
            value = decimal.Decimal(value)

        return value


class Date(Column):
    """Date column

    ::

        from datetime import date
        from anyblok.declarations import Declarations
        from anyblok.column import Date


        @Declarations.register(Declarations.Model)
        class Test:

            x = Date(default=date.today())

    """
    sqlalchemy_type = types.Date


def convert_string_to_datetime(value):
    """Convert a given value to datetime

    :param value:
    :return: datetime value
    """
    if value is None:
        return None
    elif isinstance(value, datetime):
        return value
    elif isinstance(value, date):
        return datetime.combine(value, datetime.min.time())
    elif isinstance(value, str):
        return parse(value)

    raise FieldException(
        "We can't convert this value %s to datetime")


def add_timezone_on_datetime(dt, default_timezone):
    """Convert a datetime considering the default timezone

    :param dt:
    :param default_timezone:
    :return:
    """
    if dt is not None:
        if dt.tzinfo is None:
            dt = default_timezone.localize(dt)

    return dt


class DateTimeType(types.TypeDecorator):

    impl = types.DateTime(timezone=True)
    cache_ok = True

    def __init__(self, field):
        self.default_timezone = field.default_timezone
        self.field = field

    def process_bind_param(self, value, engine):
        value = convert_string_to_datetime(value)
        value = add_timezone_on_datetime(value, self.default_timezone)
        if self.field.encrypt_key:
            return value.isoformat()

        return value

    def process_result_value(self, value, dialect):
        if self.field.encrypt_key:
            return convert_string_to_datetime(value)

        return value

    @property
    def python_type(self):
        return datetime  # pragma: no cover


class DateTime(Column):
    """DateTime column

    ::

        from datetime import datetime
        from anyblok.declarations import Declarations
        from anyblok.column import DateTime


        @Declarations.register(Declarations.Model)
        class Test:

            x = DateTime(default=datetime.now)

    """

    def __init__(self, *args, **kwargs):
        self.auto_update = kwargs.pop('auto_update', False)
        default_timezone = kwargs.pop(
            'default_timezone',
            Configuration.get('default_timezone')
        )
        if not default_timezone:
            default_timezone = time.tzname[0]

        if isinstance(default_timezone, str):
            default_timezone = pytz.timezone(default_timezone)

        self.default_timezone = default_timezone
        self.sqlalchemy_type = DateTimeType(self)
        super(DateTime, self).__init__(*args, **kwargs)

    def setter_format_value(self, value):
        """Return converted and formatted value

        :param value:
        :return:
        """
        value = convert_string_to_datetime(value)
        return add_timezone_on_datetime(value, self.default_timezone)

    def getter_format_value(self, value):
        value = convert_string_to_datetime(value)
        return add_timezone_on_datetime(value, self.default_timezone)

    def autodoc_get_properties(self):
        """Return properties for autodoc

        :return: autodoc properties
        """
        res = super(Column, self).autodoc_get_properties()
        res['is auto updated'] = self.auto_update
        if self.default_timezone:
            res['default timezone'] = self.default_timezone

        return res


class TimeStamp(DateTime):
    """ TimeStamp column

    ::

        from datetime import datetime
        from anyblok.declarations import Declarations
        from anyblok.column import DateTime


        @Declarations.register(Declarations.Model)
        class Test:

            x = TimeStamp(default=datetime.now)

    """

    def __init__(self, *args, **kwargs):
        super(TimeStamp, self).__init__(*args, **kwargs)
        self.sqlalchemy_type = types.TIMESTAMP(timezone=True)

    def getter_format_value(self, value):
        value = convert_string_to_datetime(value)
        return add_timezone_on_datetime(value, self.default_timezone)


class Time(Column):
    """Time column

    ::

        from datetime import time
        from anyblok.declarations import Declarations
        from anyblok.column import Time


        @Declarations.register(Declarations.Model)
        class Test:

            x = Time(default=time())

    """
    sqlalchemy_type = types.Time


class Interval(Column):
    """Datetime interval column

    ::

        from datetime import timedelta
        from anyblok.declarations import Declarations
        from anyblok.column import Interval


        @Declarations.register(Declarations.Model)
        class Test:

            x = Interval(default=timedelta(days=5))

    """
    sqlalchemy_type = types.Interval

    def native_type(self, registry):
        if self.encrypt_key:
            return types.VARCHAR(1024)

        return self.sqlalchemy_type

    def setter_format_value(self, value):
        if self.encrypt_key:
            value = dumps({
                x: getattr(value, x)
                for x in ['days', 'seconds', 'microseconds']})

        return value

    def getter_format_value(self, value):
        if self.encrypt_key:
            value = timedelta(**loads(value))

        return value


class StringType(types.TypeDecorator):

    impl = types.String
    cache_ok = True

    def process_bind_param(self, value, engine):
        if value is False:
            value = ''

        return value

    def process_result_value(self, value, dialect):
        return value


class String(Column):
    """String column

    ::

        from anyblok.declarations import Declarations
        from anyblok.column import String


        @Declarations.register(Declarations.Model)
        class Test:

            x = String(default='test')

    """
    def __init__(self, *args, **kwargs):
        self.size = kwargs.pop('size', 64)
        kwargs.pop('type_', None)
        self.sqlalchemy_type = StringType(self.size)
        super(String, self).__init__(*args, **kwargs)

    def autodoc_get_properties(self):
        """Return properties for autodoc

        :return: autodoc properties
        """
        res = super(String, self).autodoc_get_properties()
        res['size'] = self.size
        return res

    def get_encrypt_key_type(self, registry, sqlalchemy_type, encrypt_key):
        sqlalchemy_type = StringEncryptedType(sqlalchemy_type, encrypt_key)
        if sgdb_in(registry.engine, ['MySQL', 'MariaDB']):
            sqlalchemy_type.impl = types.String(max(self.size, 64))

        return sqlalchemy_type


class Enum(Column):
    """Enum column

    ::

        from anyblok.declarations import Declarations
        from anyblok.column import Enum
        import enum


        class MyEnumClass(enum.Enum):
            one = 1
            two = 2
            three = 3


        @Declarations.register(Declarations.Model)
        class Test:

            x = Enum(enum_cls=MyEnumClass, default='test')

    enum_cls should be an enum class
    """
    def __init__(self, *args, **kwargs):
        self.enum_cls = kwargs.pop('enum_cls')
        self.sqlalchemy_type = types.Enum(self.enum_cls)
        super(Enum, self).__init__(*args, **kwargs)

    def autodoc_get_properties(self):
        """Return properties for autodoc

        :return: autodoc properties
        """
        res = super(Enum, self).autodoc_get_properties()
        res['enum_cls'] = repr(self.enum_cls)
        return res


class MsSQLPasswordType(PasswordType):
    impl = types.VARCHAR(1024)

    def load_dialect_impl(self, dialect):
        return dialect.type_descriptor(types.VARCHAR(self.length))


class Password(Column):
    """String column

    ::

        from anyblok.declarations import Declarations
        from anyblok.column import Password


        @Declarations.register(Declarations.Model)
        class Test:

            x = Password(crypt_context={'schemes': ['md5_crypt']})

        =========================================

        test = Test.insert()
        test.x = 'mypassword'

        test.x
        ==> Password object with encrypt value, the value can not be read

        test.x == 'mypassword'
        ==> True

    ..warning::

        the column type Password can not be querying::

            Test.query().filter(Test.x == 'mypassword').count()
            ==> 0
    """
    def __init__(self, *args, **kwargs):
        self.size = kwargs.pop('size', 64)
        crypt_context = kwargs.pop('crypt_context', {})
        self.crypt_context = crypt_context
        kwargs.pop('type_', None)

        if 'foreign_key' in kwargs:
            raise FieldException("Column Password can not have a foreign key")

        self.sqlalchemy_type = PasswordType(
            max_length=self.size, **crypt_context)
        super(Password, self).__init__(*args, **kwargs)

    def setter_format_value(self, value):
        """Return formatted value

        :param value:
        :return:
        """
        value = self.sqlalchemy_type.context.hash(value).encode('utf8')
        value = SAU_PWD(value, context=self.sqlalchemy_type.context)
        return value

    def native_type(self, registry):
        """ Return the native SqlAlchemy type """
        if sgdb_in(registry.engine, ['MsSQL']):
            return MsSQLPasswordType(max_length=self.size, **self.crypt_context)

        return self.sqlalchemy_type

    def autodoc_get_properties(self):
        """Return properties for autodoc

        :return: autodoc properties
        """
        res = super(Password, self).autodoc_get_properties()
        res['size'] = self.size
        res['Crypt context'] = self.crypt_context
        return res


class TextType(types.TypeDecorator):

    impl = types.Text
    cache_ok = True

    def process_bind_param(self, value, engine):
        if value is False:
            value = ''

        return value

    def process_result_value(self, value, dialect):
        return value


class Text(Column):
    """Text column

    ::

        from anyblok.declarations import Declarations
        from anyblok.column import Text


        @Declarations.register(Declarations.Model)
        class Test:

            x = Text(default='test')

    """
    sqlalchemy_type = TextType

    def get_encrypt_key_type(self, registry, sqlalchemy_type, encrypt_key):
        sqlalchemy_type = StringEncryptedType(sqlalchemy_type, encrypt_key)
        if sgdb_in(registry.engine, ['MySQL', 'MariaDB']):
            sqlalchemy_type.impl = types.Text()

        return sqlalchemy_type


class StrSelection(str):
    """Class representing the data of one column Selection """
    selections = dumps({})
    registry = None
    namespace = None

    def get_selections(self):
        """Return a dict of selections

        :return: selections dict
        """
        selections = loads(self.selections)
        if isinstance(selections, dict):
            return selections
        if isinstance(selections, str):
            m = self.registry.get(self.namespace)
            return dict(getattr(m, selections)())

    def validate(self):
        """Validate if the key is in the selections

        :return: True or False
        """
        a = super(StrSelection, self).__str__()
        return a in self.get_selections().keys()

    @property
    def label(self):
        """Return the label corresponding to the selection key

        :return:
        """
        a = super(StrSelection, self).__str__()
        return self.get_selections()[a]


class SelectionType(types.TypeDecorator):
    """Generic type for Column Selection """

    impl = types.String
    cache_ok = True

    def __init__(self, selections, size, registry=None, namespace=None):
        super(SelectionType, self).__init__(length=size)
        self.size = size
        if isinstance(selections, (dict, str)):
            self.selections = selections
        elif isinstance(selections, (list, tuple)):
            self.selections = dict(selections)
        else:
            raise FieldException(  # pragma: no cover
                "selection wainting 'dict', get %r" % type(selections))

        if isinstance(self.selections, dict):
            for k in self.selections.keys():
                if not isinstance(k, str):
                    raise FieldException('The key must be a str')
                if len(k) > 64:
                    raise Exception(  # pragma: no cover
                        '%r is too long %r, waiting max %s or use size arg' % (
                            k, len(k), size))

        self.selections = dumps(self.selections)
        self._StrSelection = type('StrSelection', (StrSelection,),
                                  {'selections': self.selections,
                                   'registry': registry,
                                   'namespace': namespace})

    @property
    def python_type(self):
        return self._StrSelection

    def process_bind_param(self, value, engine):
        if value is not None:
            value = self.python_type(value)

        return value

    def process_result_value(self, value, dialect):
        return value


class Selection(Column):
    """Selection column

    ::

        from anyblok.declarations import Declarations
        from anyblok.column import Selection


        @Declarations.register(Declarations.Model)
        class Test:
            STATUS = (
                (u'draft', u'Draft'),
                (u'done', u'Done'),
            )

            x = Selection(selections=STATUS, size=64, default=u'draft')

    """
    def __init__(self, *args, **kwargs):
        self.selections = tuple()
        if 'selections' in kwargs:
            self.selections = kwargs.pop('selections')

        self.size = kwargs.pop('size', 64)
        self.sqlalchemy_type = 'tmp value for assert'

        super(Selection, self).__init__(*args, **kwargs)

    def autodoc_get_properties(self):
        """Return properties for autodoc

        :return: autodoc properties
        """
        res = super(Selection, self).autodoc_get_properties()
        res['selections'] = self.selections
        res['size'] = self.size
        return res

    def getter_format_value(self, value):
        """Return formatted value

        :param value:
        :return:
        """
        if value is None:
            return None

        return self.sqlalchemy_type.python_type(value)

    def setter_format_value(self, value):
        """Return value or raise exception if the given value is invalid

        :param value:
        :exception FieldException
        :return:
        """
        if value is not None:
            val = self.sqlalchemy_type.python_type(value)
            if not val.validate():
                raise FieldException('%r is not in the selections (%s)' % (
                    value, ', '.join(val.get_selections())))

        return value

    def get_sqlalchemy_mapping(self, registry, namespace, fieldname,
                               properties):
        """Return sqlalchmy mapping

        :param registry: the current registry
        :param namespace: the namespace of the model
        :param fieldname: the fieldname of the model
        :param properties: the properties of the model
        :return: instance of the real field
        """
        self.sqlalchemy_type = SelectionType(
            self.selections, self.size, registry=registry, namespace=namespace)
        return super(Selection, self).get_sqlalchemy_mapping(
            registry, namespace, fieldname, properties)

    def update_description(self, registry, model, res):
        """Update model description

        :param registry: the current registry
        :param model:
        :param res:
        """
        super(Selection, self).update_description(registry, model, res)
        sqlalchemy_type = SelectionType(
            self.selections, self.size, registry=registry, namespace=model)
        values = sqlalchemy_type._StrSelection().get_selections()
        res['selections'] = [(k, v) for k, v in values.items()]

    def must_be_copied_before_declaration(self):
        """Return True if selections is an instance of str.
        In the case of the field selection is a mixin, it must be copied or the
        selection method can fail
        """
        if isinstance(self.selections, str):
            return True
        else:
            return False

    def update_properties(self, registry, namespace, fieldname, properties):
        """Update column properties

        :param registry: the current registry
        :param namespace: the namespace of the model
        :param fieldname: the fieldname of the model
        :param properties: the properties of the model
        """
        super(Selection, self).update_properties(registry, namespace,
                                                 fieldname, properties)
        self.fieldname = fieldname
        properties['add_in_table_args'].append(self)

    def update_table_args(self, registry, Model):
        """Return check constraints to limit the value

        :param registry:
        :param Model:
        :return: list of checkConstraint
        """
        if self.encrypt_key:
            # dont add constraint because the state is crypted and nobody
            # can add new entry
            return []

        if sgdb_in(registry.engine, ['MariaDB', 'MsSQL', 'MySQL']):
            # No check constraint in MariaDB
            return []

        selections = loads(self.sqlalchemy_type.selections)
        if isinstance(selections, dict):
            enum = selections.keys()
        else:
            enum = getattr(Model, selections)()
            if isinstance(enum, (list, tuple)):
                enum = dict(enum)

            enum = enum.keys()

        if len(enum) > 1:
            constraint = """"%s" in ('%s')""" % (
                self.fieldname, "', '".join(enum))
        elif enum:
            constraint = """"%s" = '%s'""" % (self.fieldname, list(enum)[0])
        else:
            constraint = None

        if constraint:
            enum = list(enum)
            enum.sort()
            key = md5()
            key.update(str(enum).encode('utf-8'))
            name = self.fieldname + '_' + key.hexdigest() + '_types'
            return [CheckConstraint(constraint, name=name)]

        return []

    def get_encrypt_key_type(self, registry, sqlalchemy_type, encrypt_key):
        sqlalchemy_type = StringEncryptedType(sqlalchemy_type, encrypt_key)
        if sgdb_in(registry.engine, ['MySQL', 'MariaDB']):
            sqlalchemy_type.impl = types.String(max(self.size, 64))

        return sqlalchemy_type


"""
    Added *process_result_value* at the class *JSON*, because
    this method is necessary for encrypt the column
"""
types.JSON.process_result_value = lambda self, value, dialect: value


class Json(Column):
    """JSON column

    ::

        from anyblok.declarations import Declarations
        from anyblok.column import Json


        @Declarations.register(Declarations.Model)
        class Test:

            x = Json()

    """
    sqlalchemy_type = types.JSON(none_as_null=True)

    def native_type(self, registry):
        """ Return the native SqlAlchemy type """
        if sgdb_in(registry.engine, ['MariaDB', 'MsSQL']):
            return JSONType

        return self.sqlalchemy_type

    def setter_format_value(self, value):
        if self.encrypt_key:
            value = dumps(value)

        return value

    def getter_format_value(self, value):
        if value is None:
            return None

        if self.encrypt_key:
            value = loads(value)

        return value

    def get_encrypt_key_type(self, registry, sqlalchemy_type, encrypt_key):
        sqlalchemy_type = StringEncryptedType(sqlalchemy_type, encrypt_key)
        if sgdb_in(registry.engine, ['MySQL', 'MariaDB']):
            sqlalchemy_type.impl = types.Text()

        return sqlalchemy_type


class LargeBinary(Column):
    """Large binary column

    ::

        from os import urandom
        from anyblok.declarations import Declarations
        from anyblok.column import LargeBinary


        blob = urandom(100000)


        @Declarations.register(Declarations.Model)
        class Test:

            x = LargeBinary(default=blob)

    """
    sqlalchemy_type = types.LargeBinary

    def native_type(self, registry):
        if self.encrypt_key:
            return types.Text

        return self.sqlalchemy_type

    def setter_format_value(self, value):
        if self.encrypt_key:
            value = b64encode(value).decode('utf-8')

        return value

    def getter_format_value(self, value):
        if self.encrypt_key:
            value = b64decode(value.encode('utf-8'))

        return value

    def get_encrypt_key_type(self, registry, sqlalchemy_type, encrypt_key):
        sqlalchemy_type = StringEncryptedType(sqlalchemy_type, encrypt_key)
        if sgdb_in(registry.engine, ['MySQL', 'MariaDB']):
            sqlalchemy_type.impl = types.Text()

        return sqlalchemy_type


class Sequence(String):
    """Sequence column

    ::

        from anyblok.column import Sequence


        @Declarations.register(Declarations.Model)
        class Test:

            x = Sequence()

    """
    def __init__(self, *args, **kwargs):
        if 'foreign_key' in kwargs:
            raise FieldException("Sequence column can not define a foreign key"
                                 " %r" % kwargs['foreign_key'])
        if 'default' in kwargs:
            raise FieldException("Sequence column can not define a default "
                                 "value")
        kwargs['default'] = ColumnDefaultValue(self.wrap_default)

        self.code = kwargs.pop('code') if 'code' in kwargs else None
        self.start = kwargs.pop('start', 1)
        self.formater = kwargs.pop(
            'formater') if 'formater' in kwargs else None

        super(Sequence, self).__init__(*args, **kwargs)

    def autodoc_get_properties(self):
        """Return properties for autodoc

        :return: autodoc properties
        """
        res = super(Sequence, self).autodoc_get_properties()
        res['formater'] = self.formater
        return res

    def wrap_default(self, registry, namespace, fieldname, properties):
        """Return default wrapper

        :param registry: the current registry
        :param namespace: the namespace of the model
        :param fieldname: the fieldname of the model
        :return:
        """
        if not hasattr(
            registry, '_need_sequence_to_create_if_not_exist'
        ):  # pragma: no cover
            registry._need_sequence_to_create_if_not_exist = []
        elif registry._need_sequence_to_create_if_not_exist is None:
            registry._need_sequence_to_create_if_not_exist = []

        code = self.code if self.code else "%s=>%s" % (namespace, fieldname)
        registry._need_sequence_to_create_if_not_exist.append(
            {'code': code, 'formater': self.formater, 'start': self.start})

        def default_value():
            """Return next sequence value

            :return:
            """
            return registry.System.Sequence.nextvalBy(code=code)

        return default_value


class Color(Column):
    """Color column.
    `See colour package on pypi <https://pypi.python.org/pypi/colour/>`_

    ::

        from anyblok.declarations import Declarations
        from anyblok.column import Color


        @Declarations.register(Declarations.Model)
        class Test:

            x = Color(default='green')

    """
    def __init__(self, *args, **kwargs):
        self.max_length = max_length = kwargs.pop('size', 20)
        kwargs.pop('type_', None)
        self.sqlalchemy_type = ColorType(max_length)
        super(Color, self).__init__(*args, **kwargs)

    def setter_format_value(self, value):
        """Format the given value

        :param value:
        :return:
        """
        if isinstance(value, str):
            value = self.sqlalchemy_type.python_type(value)

        return value

    def autodoc_get_properties(self):
        """Return properties for autodoc

        :return: autodoc properties
        """
        res = super(Color, self).autodoc_get_properties()
        res['size'] = self.max_length
        return res

    def get_encrypt_key_type(self, registry, sqlalchemy_type, encrypt_key):
        sqlalchemy_type = StringEncryptedType(sqlalchemy_type, encrypt_key)
        if sgdb_in(registry.engine, ['MySQL', 'MariaDB']):
            sqlalchemy_type.impl = types.String(max(self.max_length, 64))

        return sqlalchemy_type


class UUID(Column):
    """UUID column

    ::

        from anyblok.column import UUID


        @Declarations.register(Declarations.Model)
        class Test:

            x = UUID()

    """
    def __init__(self, *args, **kwargs):
        uuid_kwargs = {}
        self.binary = None
        self.native = None
        for kwarg in ('binary', 'native'):
            if kwarg in kwargs:
                uuid_kwargs[kwarg] = kwargs.pop(kwarg)
                setattr(self, kwarg, uuid_kwargs[kwarg])

        self.sqlalchemy_type = UUIDType(**uuid_kwargs)
        super(UUID, self).__init__(*args, **kwargs)

    def autodoc_get_properties(self):
        """Return properties for autodoc

        :return: autodoc properties
        """
        res = super(UUID, self).autodoc_get_properties()
        res['binary'] = self.binary
        res['native'] = self.native
        return res


URLType.cache_ok = True  # waiting fix from sqlalchemy_utils


class URL(Column):
    """URL column

    ::

        from anyblok.declarations import Declarations
        from anyblok.column import URL


        @Declarations.register(Declarations.Model)
        class Test:

            x = URL(default='doc.anyblok.org')

    """
    sqlalchemy_type = URLType

    def setter_format_value(self, value):
        """Return formatted url value

        :param value:
        :return:
        """
        from furl import furl

        if value is not None:
            if isinstance(value, str):
                value = furl(value)

        return value


class PhoneNumber(Column):
    """PhoneNumber column

    ::

        from anyblok.declarations import Declarations
        from anyblok.column import PhoneNumber


        @Declarations.register(Declarations.Model)
        class Test:

            x = PhoneNumber(default='+120012301')


    .. note:: ``phonenumbers`` >= **8.9.5** distribution is required

    """
    def __init__(self, region='FR', max_length=20, *args, **kwargs):
        self.region = region
        self.max_length = max_length
        kwargs.pop('type_', None)

        self.sqlalchemy_type = PhoneNumberType(
            region=region, max_length=max_length)
        super(PhoneNumber, self).__init__(*args, **kwargs)

    def setter_format_value(self, value):
        """Return formatted phone number value

        :param value:
        :return:
        """
        if value and isinstance(value, str):
            value = self.sqlalchemy_type.python_type(value, self.region)

        return value

    def autodoc_get_properties(self):
        """Return properties for autodoc

        :return: autodoc properties
        """
        res = super(PhoneNumber, self).autodoc_get_properties()
        res['region'] = self.region
        res['max_length'] = self.max_length
        return res

    def get_encrypt_key_type(self, registry, sqlalchemy_type, encrypt_key):
        sqlalchemy_type = StringEncryptedType(sqlalchemy_type, encrypt_key)
        if sgdb_in(registry.engine, ['MySQL', 'MariaDB']):
            sqlalchemy_type.impl = types.String(max(self.max_length, 64))

        return sqlalchemy_type


"""
    Added *process_result_value* at the class *EmailType*, because
    this method is necessary for encrypt the column
"""
EmailType.process_result_value = lambda self, value, dialect: value
EmailType.cache_ok = True  # waiting fix from sqlalchemy_utils


class Email(Column):
    """Email column

    ::

        from anyblok.column import Email


        @Declarations.register(Declarations.Model)
        class Test:

            x = Email()

    """
    sqlalchemy_type = EmailType

    def setter_format_value(self, value):
        """Return formatted email value

        :param value:
        :return:
        """
        if value is not None:
            return value.lower()

        return value  # pragma: no cover


class CountryType(types.TypeDecorator, ScalarCoercible):
    """Generic type for Column Country """

    impl = types.Unicode(3)
    cache_ok = True
    python_type = python_pycountry_type

    def process_bind_param(self, value, dialect):
        if value and isinstance(value, self.python_type):
            return value.alpha_3

        return value

    def process_result_value(self, value, dialect):
        return self._coerce(value)

    def _coerce(self, value):
        if value is not None and not isinstance(value, self.python_type):
            return pycountry.countries.get(alpha_3=value)

        return value  # pragma: no cover


class Country(Column):
    """Country column.

    ::

        from anyblok.declarations import Declarations
        from anyblok.column import Country
        from pycountry import countries


        @Declarations.register(Declarations.Model)
        class Test:

            x = Country(default=countries.get(alpha_2='FR'))

    """
    sqlalchemy_type = CountryType

    def __init__(self, mode='alpha_2', *args, **kwargs):
        self.mode = mode
        if pycountry is None:
            raise FieldException(  # pragma: no cover
                "'pycountry' package is required for use 'CountryType'")

        self.choices = {getattr(country, mode): country.name
                        for country in pycountry.countries}
        super(Country, self).__init__(*args, **kwargs)

    def setter_format_value(self, value):
        """Return formatted country value

        :param value:
        :return:
        """
        if value and not isinstance(value, self.sqlalchemy_type.python_type):
            value = pycountry.countries.get(
                **{
                    self.mode: value,
                    'default': pycountry.countries.lookup(value)
                })

        return value

    def autodoc_get_properties(self):
        """Return properties for autodoc

        :return: autodoc properties
        """
        res = super(Country, self).autodoc_get_properties()
        res['mode'] = self.mode
        res['choices'] = self.choices
        return res

    def update_properties(self, registry, namespace, fieldname, properties):
        """Update column properties

        :param registry: the current registry
        :param namespace: the namespace of the model
        :param fieldname: the fieldname of the model
        :param properties: the properties of the model
        """
        super(Country, self).update_properties(registry, namespace,
                                               fieldname, properties)
        self.fieldname = fieldname
        properties['add_in_table_args'].append(self)

    def update_table_args(self, registry, Model):
        """Return check constraints to limit the value

        :param registry:
        :param Model:
        :return: list of checkConstraint
        """
        if self.encrypt_key:
            # dont add constraint because the state is crypted and nobody
            # can add new entry
            return []

        if sgdb_in(registry.engine, ['MariaDB', 'MsSQL']):
            # No Check constraint in MariaDB
            return []

        enum = [country.alpha_3 for country in pycountry.countries]
        constraint = """"%s" in ('%s')""" % (self.fieldname, "', '".join(enum))
        enum.sort()
        key = md5()
        key.update(str(enum).encode('utf-8'))
        name = self.fieldname + '_' + key.hexdigest() + '_types'
        return [CheckConstraint(constraint, name=name)]


@CompareType.add_comparator(String, String)
@CompareType.add_comparator(String, Selection)
@CompareType.add_comparator(String, Sequence)
def compare_strings(col1, type1, col2, type2):
    if type1.size != type2.size:
        raise FieldException(
            "You can't add a foreign key using based String columns with "
            "different size `{model1!s}.{col1!s}` pointing to "
            "`{model2!s}.{col2!s}` have different sizes  {type1!r}({size1:d}) "
            "-> ({type2!r}){size2:d}".format(
                model1=col1.model_name,
                col1=col1.attribute_name,
                model2=col2.model_name,
                col2=col2.attribute_name,
                type1=type1.__class__,
                type2=type2.__class__,
                size1=type1.size,
                size2=type2.size
            )
        )


@CompareType.add_comparator(String, Color)
def compare_string_to_color(col1, type1, col2, type2):
    if type1.size != type2.max_length:
        raise FieldException(
            "You can't add a foreign key using based String columns with "
            "different size `{model1!s}.{col1!s}` pointing to "
            "`{model2!s}.{col2!s}` have different sizes  {type1!r}({size1:d}) "
            "-> ({type2!r}){size2:d}".format(
                model1=col1.model_name,
                col1=col1.attribute_name,
                model2=col2.model_name,
                col2=col2.attribute_name,
                type1=type1.__class__,
                type2=type2.__class__,
                size1=type1.size,
                size2=type2.max_length
            )
        )



# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2015 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest
from anyblok.tests.testcase import LogCapture
from logging import DEBUG
from anyblok.registry import RegistryManager
from anyblok.environment import EnvironmentManager
from anyblok.model import (has_sql_fields, get_fields, ModelException,
                           has_sqlalchemy_fields)
from anyblok import Declarations
from anyblok.column import Integer, String
from .conftest import init_registry, reset_db

register = Declarations.register
unregister = Declarations.unregister
Model = Declarations.Model


class OneModel:
    __tablename__ = 'test'


class TestModel:

    @pytest.fixture(scope="class", autouse=True)
    def init_env(self, request):

        def revert():
            EnvironmentManager.set('current_blok', None)
            del RegistryManager.loaded_bloks['testModel']

        request.addfinalizer(revert)
        RegistryManager.init_blok('testModel')
        EnvironmentManager.set('current_blok', 'testModel')

    @pytest.fixture(autouse=True)
    def init_blok(self):
        blokname = 'testModel'
        RegistryManager.loaded_bloks[blokname]['Model'] = {
            'registry_names': []}

    def assertInModel(self, *args):
        blokname = 'testModel'
        blok = RegistryManager.loaded_bloks[blokname]
        assert len(blok['Model']['Model.MyModel']['bases']) == len(args)
        for cls_ in args:
            has = cls_ in blok['Model']['Model.MyModel']['bases']
            assert has is True

    def assertInRemoved(self, cls):
        core = RegistryManager.loaded_bloks['testModel']['removed']
        if cls in core:
            return True

        self.fail('Not in removed')

    def test_add_interface(self):
        register(Model, cls_=OneModel, name_='MyModel')
        assert 'Model' == Model.MyModel.__declaration_type__
        self.assertInModel(OneModel)
        dir(Declarations.Model.MyModel)

    def test_add_interface_with_decorator(self):

        @register(Model)
        class MyModel:
            pass

        assert 'Model' == Model.MyModel.__declaration_type__
        self.assertInModel(MyModel)

    def test_add_two_interface(self):

        register(Model, cls_=OneModel, name_="MyModel")

        @register(Model)
        class MyModel:
            pass

        self.assertInModel(OneModel, MyModel)

    def test_remove_interface_with_1_cls_in_registry(self):

        register(Model, cls_=OneModel, name_="MyModel")
        self.assertInModel(OneModel)
        unregister(Model.MyModel, OneModel)
        self.assertInModel(OneModel)
        self.assertInRemoved(OneModel)

    def test_remove_interface_with_2_cls_in_registry(self):

        register(Model, cls_=OneModel, name_="MyModel")

        @register(Model)
        class MyModel:
            pass

        self.assertInModel(OneModel, MyModel)
        unregister(Model.MyModel, OneModel)
        self.assertInModel(MyModel, OneModel)
        self.assertInRemoved(OneModel)


def simple_model():

    @register(Model)
    class Test:
        id = Integer(primary_key=True)
        name = String()


def simple_model_with_schema():

    @register(Model)
    class Test:
        __db_schema__ = 'test_db_schema'

        id = Integer(primary_key=True)
        name = String()


def simple_model_with_tablename():

    @register(Model, tablename='othername')
    class Test:
        id = Integer(primary_key=True)
        name = String()


def simple_models_with_same_table():

    @register(Model)
    class Test:
        id = Integer(primary_key=True)
        name = String()

    @register(Model, tablename='test')
    class Test2:
        id = Integer(primary_key=True)
        name = String()


def simple_models_with_same_table_by_declaration_model():

    @register(Model)
    class Test:
        id = Integer(primary_key=True)
        name = String()

    @register(Model, tablename=Model.Test)
    class Test2:
        id = Integer(primary_key=True)
        name = String()


def simple_models_with_same_table_by_inherit():

    @register(Model)
    class Test:
        id = Integer(primary_key=True)
        name = String()

    @register(Model)
    class Test2:
        id = Integer(primary_key=True)
        name = String()

    @register(Model, tablename='test')  # noqa
    class Test2:
        pass


def simple_models_with_inherit_sqlmodel():

    @register(Model)
    class Test:
        id = Integer(primary_key=True)
        name = String()

    @register(Model)
    class Test2(Model.Test):
        pass

    @register(Model)
    class Test3(Model.Test):
        pass


def model_with_foreign_key():

    @register(Model)
    class TestFk:

        name = String(primary_key=True)

    @register(Model)
    class Test:

        id = Integer(primary_key=True)
        name = String(foreign_key=Model.TestFk.use('name'))


class TestModel2:

    @pytest.fixture(autouse=True)
    def close_registry(self, request, bloks_loaded):

        def close():
            if hasattr(self, 'registry'):
                self.registry.close()

        request.addfinalizer(close)

    def init_registry(self, *args, **kwargs):
        reset_db()
        self.registry = init_registry(*args, **kwargs)
        return self.registry

    def test_anyblok_attribute(self):
        registry = self.init_registry(simple_model)
        self.check_registry(registry.Test)
        assert registry.System.anyblok is registry
        assert registry.System.Model.anyblok is registry
        t2 = registry.Test.query().first()
        registry.System.Model.anyblok.refresh(t2)

    def test_str(self):
        registry = self.init_registry(simple_model)
        test = registry.System.Model.query().get('Model.Test')
        assert str(test) == 'Model.Test'

    def test_deprecated_registry_attribute_getter(self):
        with pytest.warns(DeprecationWarning):
            registry = self.init_registry(simple_model)
            self.check_registry(registry.Test)
            # NoSQL Base must also have registry
            assert registry.System.registry
            assert registry.System.Model.registry
            assert registry.System.Model.registry is not registry
            t2 = registry.Test.query().first()
            registry.System.Model.registry.refresh(t2)

    def test_deprecated_registry_attribute_setter_1(self):
        with pytest.warns(DeprecationWarning):
            registry = self.init_registry(simple_model)
            self.check_registry(registry.Test)
            assert registry.System.registry.foo is None
            assert registry.System.anyblok.foo is None
            registry.System.registry.foo = 'bar'
            assert registry.System.registry.foo == 'bar'
            assert registry.System.anyblok.foo == 'bar'

    def test_deprecated_registry_attribute_setter_2(self):
        with pytest.warns(DeprecationWarning):
            registry = self.init_registry(simple_model)
            self.check_registry(registry.Test)
            assert registry.System.Model.registry.foo is None
            assert registry.System.Model.anyblok.foo is None
            registry.System.registry.foo = 'bar'
            assert registry.System.Model.registry.foo == 'bar'
            assert registry.System.Model.anyblok.foo == 'bar'

    def test_model_is_assembled(self):
        with LogCapture('anyblok.registry', level=DEBUG) as logs:
            self.init_registry(None)
            messages = logs.get_debug_messages()
            assert "Assemble 'Model' entry" in messages

    def test_model_is_initialized(self):
        with LogCapture('anyblok.registry', DEBUG) as logs:
            self.init_registry(None)
            messages = logs.get_debug_messages()
            assert "Initialize 'Model' entry" in messages

    def check_registry(self, Model):
        t = Model.insert(name="test")
        t2 = Model.query().first()
        assert t2 is t

    def check_registry_same_table(self, Model1, Model2):
        t = Model1.insert(name="test")
        t2 = Model2.query().first()
        assert t2.name == t.name

    def test_simple_model(self):
        registry = self.init_registry(simple_model)
        self.check_registry(registry.Test)
        model = registry.System.Model.query().get('Model.Test')
        assert model.table == 'test'
        assert model.schema is None

    def test_simple_model_with_schema(self):
        registry = self.init_registry(simple_model_with_schema)
        self.check_registry(registry.Test)
        model = registry.System.Model.query().get('Model.Test')
        assert model.table == 'test'
        assert model.schema == 'test_db_schema'

    def test_simple_model_with_tablename(self):
        registry = self.init_registry(simple_model_with_tablename)
        self.check_registry(registry.Test)
        assert registry.Test.__table__.name == 'othername'
        assert registry.Test.__tablename__ == 'othername'

    def test_simple_models_with_same_table(self):
        registry = self.init_registry(simple_models_with_same_table)
        self.check_registry_same_table(registry.Test, registry.Test2)

    def test_simple_models_with_same_table_by_declaration_model(self):
        registry = self.init_registry(
            simple_models_with_same_table_by_declaration_model)
        self.check_registry_same_table(registry.Test, registry.Test2)

    def test_simple_models_with_same_table_by_inherit(self):
        registry = self.init_registry(simple_models_with_same_table_by_inherit)
        self.check_registry_same_table(registry.Test, registry.Test2)

    def test_simple_models_with_inherit_sqlmodel(self):
        registry = self.init_registry(simple_models_with_inherit_sqlmodel)
        self.check_registry_same_table(registry.Test, registry.Test2)

    def test_simple_model_with_wrong_column(self):
        registry = self.init_registry(simple_model)

        try:
            registry.Test.insert(name="test", other="other")
            self.fail('No error when an inexisting colomn has filled')
        except TypeError:
            pass

    def test_model_with_foreign_key(self):
        registry = self.init_registry(model_with_foreign_key)
        registry.TestFk.insert(name='test')
        self.check_registry(registry.Test)

    def test_model_with_foreign_key_with_db_column_name(self):

        def model_with_foreign_key():

            @register(Model)
            class TestFk:

                name = String(primary_key=True, db_column_name='other')

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                name = String(foreign_key=Model.TestFk.use('name'))

        registry = self.init_registry(model_with_foreign_key)
        registry.TestFk.insert(name='test')
        self.check_registry(registry.Test)

    def test_table_args(self):
        val = 'one arg'

        def add_in_registry():

            @register(Model)
            class Test:

                @classmethod
                def define_table_args(cls):
                    return (val,)

        registry = self.init_registry(add_in_registry)
        assert len(registry.Test.__table_args__) == 1
        assert registry.Test.__table_args__[0] == val

    def test_table_args_with_inherit(self):
        val = 'first arg'
        val2 = 'second arg'

        def add_in_registry():

            @register(Model)
            class Test:

                @classmethod
                def define_table_args(cls):
                    return (val,)

            @register(Model)  # noqa
            class Test:

                @classmethod
                def define_table_args(cls):
                    return super(Test, cls).define_table_args() + (val2,)

        registry = self.init_registry(add_in_registry)
        assert len(registry.Test.__table_args__) == 2
        assert val in registry.Test.__table_args__
        assert val2 in registry.Test.__table_args__

    def test_table_args_in_cls_attribute(self):

        def add_in_registry():

            @register(Model)
            class Test:
                __table_args__ = ()

        with pytest.raises(ModelException):
            self.init_registry(add_in_registry)

    def test_mapper_args(self):
        val = 'one arg'

        def add_in_registry():

            @register(Model)
            class Test:

                @classmethod
                def define_mapper_args(cls):
                    return {val: val}

        registry = self.init_registry(add_in_registry)
        assert len(registry.Test.__mapper_args__) == 1
        assert registry.Test.__mapper_args__[val] == val

    def test_mapper_args_with_inherit(self):
        val = 'first arg'
        val2 = 'second arg'

        def add_in_registry():

            @register(Model)
            class Test:

                @classmethod
                def define_mapper_args(cls):
                    return {val: val}

            @register(Model)  # noqa
            class Test:

                @classmethod
                def define_mapper_args(cls):
                    mapper_args = super(Test, cls).define_mapper_args()
                    mapper_args.update({val2: val2})
                    return mapper_args

        registry = self.init_registry(add_in_registry)
        assert len(registry.Test.__mapper_args__) == 2
        assert registry.Test.__mapper_args__[val] == val
        assert registry.Test.__mapper_args__[val2] == val2

    def test_mapper_args_in_cls_attribute(self):

        def add_in_registry():

            @register(Model)
            class Test:
                __mapper_args__ = {}

        with pytest.raises(ModelException):
            self.init_registry(add_in_registry)

    def test_with_sqlalchemy_fields(self):
        from sqlalchemy import Column as SaC, String as SaS

        def add_in_registry():

            @register(Model)
            class Test:
                one_field = SaC(SaS(64))

        with pytest.raises(ModelException):
            self.init_registry(add_in_registry)

    def test_update_0(self):
        registry = self.init_registry(None)
        t = registry.System.Blok.query().first()
        t.update(**{'state': 'undefined'})
        assert t.state == 'undefined'

    def test_update_1(self):
        registry = self.init_registry(None)
        t = registry.System.Blok.query().first()
        t.update(state='undefined')
        assert t.state == 'undefined'

    def test_fix_issue_52(self):

        def add_in_registry():

            @register(Model)
            class Test:
                id = Integer(primary_key=True)
                type = String()

        registry = self.init_registry(add_in_registry)
        assert registry.Test is not None


class TestModelAssembling:

    def test_has_sql_fields_ok(self):

        class MyModel:
            one_field = String()

        assert has_sql_fields([MyModel]) is True

    def test_has_sql_fields_ko(self):

        class MyModel:
            one_field = None

        assert has_sql_fields([MyModel]) is False

    def test_get_fields(self):

        class MyModel:
            one_field = String()

        assert get_fields(MyModel) == {'one_field': MyModel.one_field}

    def test_has_sqlalchemy_fields(self):
        from sqlalchemy import Column as SaC, String as SaS

        class MyModel:
            one_field = SaC(SaS(64))

        assert has_sqlalchemy_fields(MyModel)

    def test_has_sqlalchemy_fields2(self):

        class MyModel:
            one_field = String()

        assert not (has_sqlalchemy_fields(MyModel))



# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2015 Pierre Verkest <pverkest@anybox.fr>
#    Copyright (C) 2019 Jean-Sebastien SUZANNE <js.suzanne@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from anyblok.testing import sgdb_in
import pytest
from anyblok import Declarations
from anyblok.config import Configuration
from anyblok.mapper import ModelAttributeException
from anyblok.column import Integer, String, DateTime
from anyblok.relationship import Many2Many, Many2One, ordering_list
from anyblok.field import FieldException
from datetime import datetime
from .conftest import init_registry, reset_db

register = Declarations.register
Model = Declarations.Model
Mixin = Declarations.Mixin


@pytest.fixture(
    scope="class",
    params=[
        ('prefix', 'suffix'),
        ('', ''),
    ]
)
def db_schema(request, bloks_loaded):
    Configuration.set('prefix_db_schema', request.param[0])
    Configuration.set('suffix_db_schema', request.param[1])

    def rollback():
        Configuration.set('prefix_db_schema', '')
        Configuration.set('suffix_db_schema', '')

    request.addfinalizer(rollback)


def _complete_many2many(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:

        name = String(primary_key=True)
        addresses = Many2Many(model=Model.Address,
                              join_table="join_addresses_by_persons",
                              remote_columns="id", local_columns="name",
                              m2m_remote_columns='a_id',
                              m2m_local_columns='p_name',
                              many2many="persons")


def _complete_many2many_with_ordered_backref(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:

        name = String(primary_key=True)
        addresses = Many2Many(model=Model.Address,
                              join_table="join_addresses_by_persons",
                              remote_columns="id", local_columns="name",
                              m2m_remote_columns='a_id',
                              m2m_local_columns='p_name',
                              many2many=(
                                  "persons",
                                  dict(
                                      order_by="ModelPerson.name",
                                      collection_class=ordering_list('name'),
                                  ),
                              ))


def _complete_many2many_with_schema(**kwargs):

    @register(Model)
    class Address:
        __db_schema__ = 'test_db_m2m_schema'

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:
        __db_schema__ = 'test_db_m2m_schema'

        name = String(primary_key=True)
        addresses = Many2Many(model=Model.Address,
                              join_table="join_addresses_by_persons",
                              schema="test_db_m2m_schema",
                              remote_columns="id", local_columns="name",
                              m2m_remote_columns='a_id',
                              m2m_local_columns='p_name',
                              many2many="persons")


def _complete_many2many_with_diferent_schema1(**kwargs):

    @register(Model)
    class Address:
        __db_schema__ = 'test_db_m2m_schema'

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:
        __db_schema__ = 'test_db_m2m_other_schema'

        name = String(primary_key=True)
        addresses = Many2Many(model=Model.Address,
                              join_table="join_addresses_by_persons",
                              remote_columns="id", local_columns="name",
                              m2m_remote_columns='a_id',
                              m2m_local_columns='p_name',
                              many2many="persons")


def _complete_many2many_with_diferent_schema2(**kwargs):

    @register(Model)
    class Address:
        __db_schema__ = 'test_db_m2m_schema'

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:

        name = String(primary_key=True)
        addresses = Many2Many(model=Model.Address,
                              join_table="join_addresses_by_persons",
                              remote_columns="id", local_columns="name",
                              m2m_remote_columns='a_id',
                              m2m_local_columns='p_name',
                              many2many="persons")


@pytest.fixture(
    scope="class",
    params=[
        _complete_many2many,
        _complete_many2many_with_ordered_backref,
        _complete_many2many_with_schema,
        _complete_many2many_with_diferent_schema1,
        _complete_many2many_with_diferent_schema2,
    ]
)
def registry_many2many(request, bloks_loaded, db_schema):
    reset_db()
    registry = init_registry(request.param)
    request.addfinalizer(registry.close)
    return registry


@pytest.mark.relationship
class TestMany2ManyComplete:

    @pytest.fixture(autouse=True)
    def transact(self, request, registry_many2many):
        transaction = registry_many2many.begin_nested()
        request.addfinalizer(transaction.rollback)
        return

    def test_complete_many2many(self, registry_many2many):
        registry = registry_many2many

        address_exist = hasattr(registry.Person, 'addresses')
        assert address_exist

        m2m_tables_exist = hasattr(registry, 'many2many_tables')
        assert m2m_tables_exist

        jt = registry.declarativebase.metadata.tables
        join_table = 'join_addresses_by_persons'
        if registry.Person.__db_schema__:
            join_table = '%s.%s' % (registry.Person.__db_schema__, join_table)
        assert join_table in jt

        assert len(jt[join_table].primary_key.columns) == 2
        assert 'a_id' in jt[join_table].primary_key.columns
        assert 'p_name' in jt[join_table].primary_key.columns

        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')

        person = registry.Person.insert(name="Jean-sébastien SUZANNE")

        person.addresses.append(address)

        assert address.persons == [person]

    def test_many2many_autodoc(self, registry_many2many):
        registry = registry_many2many
        registry.loaded_namespaces_first_step['Model.Person'][
            'addresses'].autodoc_get_properties()

    def test_complete_many2many_expire_field(self, registry_many2many):
        registry = registry_many2many
        assert ('x2m', 'addresses', 'persons') in registry.expire_attributes[
            'Model.Person']['addresses']
        assert ('x2m', 'persons', 'addresses') in registry.expire_attributes[
            'Model.Address']['persons']


def _minimum_many2many(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:

        name = String(primary_key=True)
        addresses = Many2Many(model=Model.Address)


def many2many_on_mapper(**kwargs):

    @register(Model, tablename="x")
    class Address:

        id = Integer(primary_key=True, db_column_name="x1")
        street = String(db_column_name="x2")
        zip = String(db_column_name="x3")
        city = String(db_column_name="x4")

    @register(Model, tablename="y")
    class Person:

        name = String(primary_key=True, db_column_name="y1")
        addresses = Many2Many(model=Model.Address)


def _minimum_many2many_with_schema(**kwargs):

    @register(Model)
    class Address:
        __db_schema__ = 'test_db_m2m_schema'

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:
        __db_schema__ = 'test_db_m2m_schema'

        name = String(primary_key=True)
        addresses = Many2Many(model=Model.Address)


def _minimum_many2many_by_mixin(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Mixin)
    class MixinM2M:
        addresses = Many2Many(model=Model.Address)

    @register(Model)
    class Person(Mixin.MixinM2M):

        name = String(primary_key=True)


def _many2many_with_str_model(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        street = String()
        zip = String()
        city = String()

    @register(Model)
    class Person:

        name = String(primary_key=True)
        addresses = Many2Many(model='Model.Address')


def auto_detect_two_primary_keys(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)
        id2 = Integer(primary_key=True, autoincrement=False)

    @register(Model)
    class Person:

        name = String(primary_key=True)
        addresses = Many2Many(model=Model.Address)


def unexisting_remote_columns(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)

    @register(Model)
    class Person:

        name = String(primary_key=True)
        addresses = Many2Many(model=Model.Address, remote_columns="id2")


def reuse_many2many_table(**kwargs):

    @register(Model)
    class Address:

        id = Integer(primary_key=True)

    @register(Model)
    class Person:

        name = String(primary_key=True)
        addresses = Many2Many(model=Model.Address)

    @register(Model)  # noqa
    class Address:

        id = Integer(primary_key=True)
        persons = Many2Many(
            model=Model.Person,
            join_table='join_person_and_address_for_addresses')


@pytest.mark.relationship
class TestMany2Many:

    @pytest.fixture(autouse=True)
    def close_registry(self, request, bloks_loaded):

        def close():
            if hasattr(self, 'registry'):
                self.registry.close()

        request.addfinalizer(close)

    def init_registry(self, *args, **kwargs):
        reset_db()
        self.registry = init_registry(*args, **kwargs)
        return self.registry

    def test_minimum_many2many(self):
        registry = self.init_registry(_minimum_many2many)

        address_exist = hasattr(registry.Person, 'addresses')
        assert address_exist

        m2m_tables_exist = hasattr(registry, 'many2many_tables')
        assert m2m_tables_exist

        jt = registry.declarativebase.metadata.tables
        join_table_exist = 'join_person_and_address_for_addresses' in jt
        assert join_table_exist

        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')

        person = registry.Person.insert(name="Jean-sébastien SUZANNE")

        person.addresses.append(address)

        assert person.addresses == [address]

    def test_many2many_on_mapper(self):
        registry = self.init_registry(many2many_on_mapper)

        address_exist = hasattr(registry.Person, 'addresses')
        assert address_exist

        m2m_tables_exist = hasattr(registry, 'many2many_tables')
        assert m2m_tables_exist

        jt = registry.declarativebase.metadata.tables
        join_table_exist = 'join_y_and_x_for_addresses' in jt
        assert join_table_exist

        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')

        person = registry.Person.insert(name="Jean-sébastien SUZANNE")

        person.addresses.append(address)

        assert person.addresses == [address]

    def test_minimum_many2many_with_schema(self, db_schema):
        registry = self.init_registry(_minimum_many2many_with_schema)

        address_exist = hasattr(registry.Person, 'addresses')
        assert address_exist

        m2m_tables_exist = hasattr(registry, 'many2many_tables')
        assert m2m_tables_exist

        jt = registry.declarativebase.metadata.tables
        join_table_exist = (
            '%s.join_person_and_address_for_addresses' % (
                registry.Person.__db_schema__
            ) in jt)
        assert join_table_exist

        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')

        person = registry.Person.insert(name="Jean-sébastien SUZANNE")

        person.addresses.append(address)

        assert person.addresses == [address]

    def test_many2many_with_str_model(self):
        registry = self.init_registry(_many2many_with_str_model)

        address_exist = hasattr(registry.Person, 'addresses')
        assert address_exist

        m2m_tables_exist = hasattr(registry, 'many2many_tables')
        assert m2m_tables_exist

        jt = registry.declarativebase.metadata.tables
        join_table_exist = 'join_person_and_address_for_addresses' in jt
        assert join_table_exist

        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')

        person = registry.Person.insert(name="Jean-sébastien SUZANNE")

        person.addresses.append(address)

        assert person.addresses == [address]

    def test_many2many_with_two_primary_key(self):
        registry = self.init_registry(auto_detect_two_primary_keys)

        address = registry.Address.insert(id2=1)
        person = registry.Person.insert(name="Jean-sébastien SUZANNE")
        person.addresses.append(address)

        assert person.addresses == [address]

    def test_unexisting_remote_columns(self):
        with pytest.raises(ModelAttributeException):
            self.init_registry(unexisting_remote_columns)

    def test_reuse_many2many_table(self):
        registry = self.init_registry(reuse_many2many_table)

        address = registry.Address.insert()
        person = registry.Person.insert(name="Jean-sébastien SUZANNE")

        person.addresses.append(address)

        assert address.persons == [person]

    def test_declared_in_mixin(self):
        registry = self.init_registry(_minimum_many2many_by_mixin)

        address_exist = hasattr(registry.Person, 'addresses')
        assert address_exist

        m2m_tables_exist = hasattr(registry, 'many2many_tables')
        assert m2m_tables_exist

        jt = registry.declarativebase.metadata.tables
        join_table_exist = 'join_person_and_address_for_addresses' in jt
        assert join_table_exist

        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')

        person = registry.Person.insert(name="Jean-sébastien SUZANNE")

        person.addresses.append(address)

        assert person.addresses == [address]

    @pytest.mark.skipif(sgdb_in(['MySQL', 'MariaDB', 'MsSQL']),
                        reason='ISSUE #90')
    def test_declared_in_mixin_inherit_by_two_models(self):
        def add_in_registry():
            _minimum_many2many_by_mixin()

            @register(Model)
            class Person2(Mixin.MixinM2M):

                name = String(primary_key=True)

        registry = self.init_registry(add_in_registry)

        address_exist = hasattr(registry.Person, 'addresses')
        assert address_exist
        address_exist = hasattr(registry.Person2, 'addresses')
        assert address_exist

        m2m_tables_exist = hasattr(registry, 'many2many_tables')
        assert m2m_tables_exist

        jt = registry.declarativebase.metadata.tables
        join_table_exist = 'join_person_and_address_for_addresses' in jt
        assert join_table_exist
        join_table_exist = 'join_person2_and_address_for_addresses' in jt
        assert join_table_exist

    def test_comlet_with_multi_primary_keys_remote_and_local(self):

        def add_in_registry():

            @register(Model)
            class Test:
                id = Integer(primary_key=True)
                id2 = String(primary_key=True)

            @register(Model)
            class Test2:
                id = Integer(primary_key=True)
                id2 = String(primary_key=True)
                test = Many2Many(model=Model.Test,
                                 join_table="join_test_and_test2",
                                 remote_columns=['id', 'id2'],
                                 m2m_remote_columns=['t2_id', 't2_id2'],
                                 local_columns=['id', 'id2'],
                                 m2m_local_columns=['t1_id', 't1_id2'],
                                 many2many="test2")

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert(id2="test1")
        t2 = registry.Test2.insert(id2="test2")
        t2.test.append(t1)
        assert t1.test2[0] is t2

        jt = registry.declarativebase.metadata.tables
        assert len(jt['join_test_and_test2'].primary_key.columns) == 4
        assert 't1_id' in jt[
            'join_test_and_test2'].primary_key.columns
        assert 't1_id2' in jt[
            'join_test_and_test2'].primary_key.columns
        assert 't2_id' in jt[
            'join_test_and_test2'].primary_key.columns
        assert 't2_id2' in jt[
            'join_test_and_test2'].primary_key.columns

    def test_comlet_with_multi_primary_keys_remote(self):

        def add_in_registry():

            @register(Model)
            class Test:
                id = Integer(primary_key=True)
                id2 = String(primary_key=True)

            @register(Model)
            class Test2:
                id = Integer(primary_key=True)
                test = Many2Many(model=Model.Test,
                                 join_table="join_test_and_test2",
                                 remote_columns=['id', 'id2'],
                                 m2m_remote_columns=['t2_id', 't2_id2'],
                                 local_columns='id',
                                 m2m_local_columns='t1_id',
                                 many2many="test2")

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert(id2="test1")
        t2 = registry.Test2.insert()
        t2.test.append(t1)
        assert t1.test2[0] is t2

    def test_comlet_with_multi_primary_keys_local(self):

        def add_in_registry():

            @register(Model)
            class Test:
                id = Integer(primary_key=True)

            @register(Model)
            class Test2:
                id = Integer(primary_key=True)
                id2 = String(primary_key=True)
                test = Many2Many(model=Model.Test,
                                 join_table="join_test_and_test2",
                                 remote_columns='id',
                                 m2m_remote_columns='t2_id',
                                 local_columns=['id', 'id2'],
                                 m2m_local_columns=['t1_id', 't1_id2'],
                                 many2many="test2")

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test2.insert(id2='test2')
        t2.test.append(t1)
        assert t1.test2[0] is t2

    def test_minimum_with_multi_primary_keys_remote_and_local(self):

        def add_in_registry():

            @register(Model)
            class Test:
                id = Integer(primary_key=True)
                id2 = String(primary_key=True)

            @register(Model)
            class Test2:
                id = Integer(primary_key=True)
                id2 = String(primary_key=True)
                test = Many2Many(model=Model.Test, many2many="test2")

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert(id2="test1")
        t2 = registry.Test2.insert(id2="test2")
        t2.test.append(t1)
        assert t1.test2[0] is t2

    def test_minimum_with_multi_primary_keys_remote(self):

        def add_in_registry():

            @register(Model)
            class Test:
                id = Integer(primary_key=True)
                id2 = String(primary_key=True)

            @register(Model)
            class Test2:
                id = Integer(primary_key=True)
                test = Many2Many(model=Model.Test, many2many="test2")

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert(id2="test1")
        t2 = registry.Test2.insert()
        t2.test.append(t1)
        assert t1.test2[0] is t2

    def test_minimum_with_multi_primary_keys_local(self):

        def add_in_registry():

            @register(Model)
            class Test:
                id = Integer(primary_key=True)

            @register(Model)
            class Test2:
                id = Integer(primary_key=True)
                id2 = String(primary_key=True)
                test = Many2Many(model=Model.Test, many2many="test2")

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test2.insert(id2="test2")
        t2.test.append(t1)
        assert t1.test2[0] is t2

    def test_many2many_on_self(self):

        def add_in_registry():

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                childs = Many2Many(
                    model='Model.Test',
                    m2m_remote_columns='id2',
                    many2many='parents'
                )

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test.insert()
        t1.childs.append(t2)
        assert t1 in t2.parents

    def test_many2many_on_self_with_schema(self, db_schema):

        def add_in_registry():

            @register(Model)
            class Test:
                __db_schema__ = 'test_db_m2m_schema'

                id = Integer(primary_key=True)
                childs = Many2Many(
                    model='Model.Test',
                    m2m_remote_columns='id2',
                    many2many='parents'
                )

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test.insert()
        t1.childs.append(t2)
        assert t1 in t2.parents

    def test_many2many_on_self_auto_column(self):

        def add_in_registry():

            @register(Model)
            class Test2:

                id = Integer(primary_key=True)
                childs = Many2Many(
                    model='Model.Test2',
                    many2many='parents'
                )

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test2.insert()
        t2 = registry.Test2.insert()
        t1.childs.append(t2)
        assert t1 in t2.parents

    def test_many2many_on_self_auto_column_with_schema(self, db_schema):

        def add_in_registry():

            @register(Model)
            class Test2:
                __db_schema__ = 'test_db_m2m_schema'

                id = Integer(primary_key=True)
                childs = Many2Many(
                    model='Model.Test2',
                    many2many='parents'
                )

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test2.insert()
        t2 = registry.Test2.insert()
        t1.childs.append(t2)
        assert t1 in t2.parents

    def test_many2many_on_multi_fk_miss_on_m2m_column(self):

        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)
                street = String(primary_key=True)

            @register(Model)
            class Person:

                name = String(primary_key=True)
                addresses = Many2Many(
                    m2m_remote_columns=['test_id'],
                    model=Model.Address)

        with pytest.raises(FieldException):
            self.init_registry(add_in_registry)

    def assert_rich_many2many_complete_config(self, registry):
        person = registry.Person.insert(name='jssuzanne')
        address = registry.Address.insert(
            street='somewhere', zip="75001", city="Paris")
        person.addresses.append(address)
        personaddress = registry.PersonAddress.query().one()
        assert personaddress.a_id == address.id
        assert personaddress.p_name == person.name
        assert personaddress.id
        assert personaddress.create_at
        assert personaddress.foo == 'bar'

    def test_rich_many2many_complete_config(self):
        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)
                street = String()
                zip = String()
                city = String()

            @register(Model)
            class PersonAddress:
                id = Integer(primary_key=True)
                a_id = Integer(
                    foreign_key=Model.Address.use('id'), nullable=False)
                p_name = String(
                    foreign_key='Model.Person=>name', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Person:

                name = String(primary_key=True)
                addresses = Many2Many(model=Model.Address,
                                      join_table="personaddress",
                                      remote_columns="id", local_columns="name",
                                      m2m_remote_columns='a_id',
                                      m2m_local_columns='p_name',
                                      many2many="persons")

        registry = self.init_registry(add_in_registry)
        self.assert_rich_many2many_complete_config(registry)

    def test_rich_many2many_complete_config_on_mapper(self):
        def add_in_registry():

            @register(Model, tablename="x")
            class Address:

                id = Integer(primary_key=True, db_column_name="x1")
                street = String(db_column_name="x2")
                zip = String(db_column_name="x3")
                city = String(db_column_name="x4")

            @register(Model, tablename="y")
            class PersonAddress:
                id = Integer(primary_key=True, db_column_name="y1")
                a_id = Integer(
                    foreign_key=Model.Address.use('id'), nullable=False,
                    db_column_name="y2")
                p_name = String(
                    foreign_key='Model.Person=>name', nullable=False,
                    db_column_name="y3")
                create_at = DateTime(default=datetime.now, db_column_name="y4")
                foo = String(default='bar', db_column_name="y5")

            @register(Model, tablename="z")
            class Person:

                name = String(primary_key=True, db_column_name="z1")
                addresses = Many2Many(model=Model.Address,
                                      join_model=Model.PersonAddress,
                                      many2many="persons")

        registry = self.init_registry(add_in_registry)
        self.assert_rich_many2many_complete_config(registry)

    def test_rich_many2many_complete_config_with_schema(self, db_schema):
        def add_in_registry():

            @register(Model)
            class Address:
                __db_schema__ = 'test_db_m2m_schema'

                id = Integer(primary_key=True)
                street = String()
                zip = String()
                city = String()

            @register(Model)
            class PersonAddress:
                __db_schema__ = 'test_db_m2m_schema'

                id = Integer(primary_key=True)
                a_id = Integer(
                    foreign_key=Model.Address.use('id'), nullable=False)
                p_name = String(
                    foreign_key='Model.Person=>name', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Person:
                __db_schema__ = 'test_db_m2m_schema'

                name = String(primary_key=True)
                addresses = Many2Many(
                    model=Model.Address,
                    join_table="test_db_m2m_schema.personaddress",
                    remote_columns="id", local_columns="name",
                    m2m_remote_columns='a_id',
                    m2m_local_columns='p_name',
                    many2many="persons")

        registry = self.init_registry(add_in_registry)
        self.assert_rich_many2many_complete_config(registry)

    def test_rich_many2many_complete_config_with_different_schema(self,
                                                                  db_schema):
        def add_in_registry():

            @register(Model)
            class Address:
                __db_schema__ = 'test_db_m2m_schema1'

                id = Integer(primary_key=True)
                street = String()
                zip = String()
                city = String()

            @register(Model)
            class PersonAddress:
                __db_schema__ = 'test_db_m2m_schema2'

                id = Integer(primary_key=True)
                a_id = Integer(
                    foreign_key=Model.Address.use('id'), nullable=False)
                p_name = String(
                    foreign_key='Model.Person=>name', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Person:
                __db_schema__ = 'test_db_m2m_schema3'

                name = String(primary_key=True)
                addresses = Many2Many(
                    model=Model.Address,
                    join_table="test_db_m2m_schema2.personaddress",
                    remote_columns="id", local_columns="name",
                    m2m_remote_columns='a_id',
                    m2m_local_columns='p_name',
                    many2many="persons")

        registry = self.init_registry(add_in_registry)
        self.assert_rich_many2many_complete_config(registry)

    def test_rich_many2many_minimum_config(self):

        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)
                street = String()
                zip = String()
                city = String()

            @register(Model)
            class PersonAddress:
                id = Integer(primary_key=True)
                a_id = Integer(
                    foreign_key=Model.Address.use('id'), nullable=False)
                p_name = String(
                    foreign_key='Model.Person=>name', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Person:

                name = String(primary_key=True)
                addresses = Many2Many(model=Model.Address,
                                      join_table="personaddress",
                                      many2many="persons")

        registry = self.init_registry(add_in_registry)
        person = registry.Person.insert(name='jssuzanne')
        address = registry.Address.insert(
            street='somewhere', zip="75001", city="Paris")
        person.addresses.append(address)
        personaddress = registry.PersonAddress.query().one()
        assert personaddress.a_id == address.id
        assert personaddress.p_name == person.name
        assert personaddress.id
        assert personaddress.create_at
        assert personaddress.foo == 'bar'

    def test_rich_many2many_minimum_config_on_join_model(self):

        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)
                street = String()
                zip = String()
                city = String()

            @register(Model)
            class PersonAddress:
                id = Integer(primary_key=True)
                a_id = Integer(
                    foreign_key=Model.Address.use('id'), nullable=False)
                p_name = String(
                    foreign_key='Model.Person=>name', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Person:

                name = String(primary_key=True)
                addresses = Many2Many(model=Model.Address,
                                      join_model=Model.PersonAddress,
                                      many2many="persons")

        registry = self.init_registry(add_in_registry)
        person = registry.Person.insert(name='jssuzanne')
        address = registry.Address.insert(
            street='somewhere', zip="75001", city="Paris")
        person.addresses.append(address)
        personaddress = registry.PersonAddress.query().one()
        assert personaddress.a_id == address.id
        assert personaddress.p_name == person.name
        assert personaddress.id
        assert personaddress.create_at
        assert personaddress.foo == 'bar'

    def test_rich_many2many_minimum_config_on_join_model_with_schema(self,
                                                                     db_schema):

        def add_in_registry():

            @register(Model)
            class Address:
                __db_schema__ = 'test_db_m2m_schema'

                id = Integer(primary_key=True)
                street = String()
                zip = String()
                city = String()

            @register(Model)
            class PersonAddress:
                __db_schema__ = 'test_db_m2m_schema'

                id = Integer(primary_key=True)
                a_id = Integer(
                    foreign_key=Model.Address.use('id'), nullable=False)
                p_name = String(
                    foreign_key='Model.Person=>name', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Person:
                __db_schema__ = 'test_db_m2m_schema'

                name = String(primary_key=True)
                addresses = Many2Many(model=Model.Address,
                                      join_model=Model.PersonAddress,
                                      many2many="persons")

        registry = self.init_registry(add_in_registry)
        person = registry.Person.insert(name='jssuzanne')
        address = registry.Address.insert(
            street='somewhere', zip="75001", city="Paris")
        person.addresses.append(address)
        personaddress = registry.PersonAddress.query().one()
        assert personaddress.a_id == address.id
        assert personaddress.p_name == person.name
        assert personaddress.id
        assert personaddress.create_at
        assert personaddress.foo == 'bar'

    def test_rich_many2many_minimum_config_on_join_model_with_dif_schema(
        self, db_schema
    ):

        def add_in_registry():

            @register(Model)
            class Address:
                __db_schema__ = 'test_db_m2m_schema1'

                id = Integer(primary_key=True)
                street = String()
                zip = String()
                city = String()

            @register(Model)
            class PersonAddress:
                __db_schema__ = 'test_db_m2m_schema2'

                id = Integer(primary_key=True)
                a_id = Integer(
                    foreign_key=Model.Address.use('id'), nullable=False)
                p_name = String(
                    foreign_key='Model.Person=>name', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Person:
                __db_schema__ = 'test_db_m2m_schema3'

                name = String(primary_key=True)
                addresses = Many2Many(model=Model.Address,
                                      join_model=Model.PersonAddress,
                                      many2many="persons")

        registry = self.init_registry(add_in_registry)
        person = registry.Person.insert(name='jssuzanne')
        address = registry.Address.insert(
            street='somewhere', zip="75001", city="Paris")
        person.addresses.append(address)
        personaddress = registry.PersonAddress.query().one()
        assert personaddress.a_id == address.id
        assert personaddress.p_name == person.name
        assert personaddress.id
        assert personaddress.create_at
        assert personaddress.foo == 'bar'

    def test_rich_many2many_minimum_config_on_join_model_with_di_schema2(
        self, db_schema
    ):

        def add_in_registry():

            @register(Model)
            class Address:
                __db_schema__ = 'test_db_m2m_schema1'

                id = Integer(primary_key=True)
                street = String()
                zip = String()
                city = String()

            @register(Model)
            class PersonAddress:

                id = Integer(primary_key=True)
                a_id = Integer(
                    foreign_key=Model.Address.use('id'), nullable=False)
                p_name = String(
                    foreign_key='Model.Person=>name', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Person:
                __db_schema__ = 'test_db_m2m_schema3'

                name = String(primary_key=True)
                addresses = Many2Many(model=Model.Address,
                                      join_model=Model.PersonAddress,
                                      many2many="persons")

        registry = self.init_registry(add_in_registry)
        person = registry.Person.insert(name='jssuzanne')
        address = registry.Address.insert(
            street='somewhere', zip="75001", city="Paris")
        person.addresses.append(address)
        personaddress = registry.PersonAddress.query().one()
        assert personaddress.a_id == address.id
        assert personaddress.p_name == person.name
        assert personaddress.id
        assert personaddress.create_at
        assert personaddress.foo == 'bar'

    def test_rich_many2many_minimum_config_with_many2one(self):

        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)
                street = String()
                zip = String()
                city = String()

            @register(Model)
            class PersonAddress:
                id = Integer(primary_key=True)
                person = Many2One(
                    model='Model.Person', nullable=False,
                    foreign_key_options={'ondelete': 'cascade'})
                address = Many2One(
                    model=Model.Address, nullable=False,
                    foreign_key_options={'ondelete': 'cascade'})
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Person:

                name = String(primary_key=True)
                addresses = Many2Many(model=Model.Address,
                                      join_model=Model.PersonAddress,
                                      many2many="persons")

        registry = self.init_registry(add_in_registry)
        person = registry.Person.insert(name='jssuzanne')
        address = registry.Address.insert(
            street='somewhere', zip="75001", city="Paris")
        person.addresses.append(address)
        personaddress = registry.PersonAddress.query().one()
        assert personaddress.address_id == address.id
        assert personaddress.person_name == person.name
        assert personaddress.id
        assert personaddress.create_at
        assert personaddress.foo == 'bar'

    def test_rich_many2many_minimum_config_with_pk_many2one(self):

        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)
                street = String()
                zip = String()
                city = String()

            @register(Model)
            class PersonAddress:
                person = Many2One(
                    model='Model.Person', nullable=False,
                    primary_key=True,
                    foreign_key_options={'ondelete': 'cascade'})
                address = Many2One(
                    model=Model.Address, nullable=False,
                    primary_key=True,
                    foreign_key_options={'ondelete': 'cascade'})
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Person:

                name = String(primary_key=True)
                addresses = Many2Many(model=Model.Address,
                                      join_model=Model.PersonAddress,
                                      many2many="persons")

        registry = self.init_registry(add_in_registry)
        person = registry.Person.insert(name='jssuzanne')
        address = registry.Address.insert(
            street='somewhere', zip="75001", city="Paris")
        person.addresses.append(address)
        personaddress = registry.PersonAddress.query().one()
        assert personaddress.address_id == address.id
        assert personaddress.person_name == person.name
        assert personaddress.create_at
        assert personaddress.foo == 'bar'

    def test_rich_many2many_minimum_config_on_join_model_and_join_table_1(self):

        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)
                street = String()
                zip = String()
                city = String()

            @register(Model)
            class PersonAddress:
                id = Integer(primary_key=True)
                a_id = Integer(
                    foreign_key=Model.Address.use('id'), nullable=False)
                p_name = String(
                    foreign_key='Model.Person=>name', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Person:

                name = String(primary_key=True)
                addresses = Many2Many(model=Model.Address,
                                      join_model=Model.PersonAddress,
                                      many2many="persons")

        registry = self.init_registry(add_in_registry)
        person = registry.Person.insert(name='jssuzanne')
        address = registry.Address.insert(
            street='somewhere', zip="75001", city="Paris")
        person.addresses.append(address)
        personaddress = registry.PersonAddress.query().one()
        assert personaddress.a_id == address.id
        assert personaddress.p_name == person.name
        assert personaddress.id
        assert personaddress.create_at
        assert personaddress.foo == 'bar'

    def test_rich_many2many_minimum_config_on_join_model_and_join_table_2(self):

        def add_in_registry():

            @register(Model)
            class Address:

                id = Integer(primary_key=True)
                street = String()
                zip = String()
                city = String()

            @register(Model)
            class PersonAddress:
                id = Integer(primary_key=True)
                a_id = Integer(
                    foreign_key=Model.Address.use('id'), nullable=False)
                p_name = String(
                    foreign_key='Model.Person=>name', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Person:

                name = String(primary_key=True)
                addresses = Many2Many(model=Model.Address,
                                      join_table="other_join_table",
                                      join_model=Model.PersonAddress,
                                      many2many="persons")

        with pytest.raises(FieldException):
            self.init_registry(add_in_registry)

    def test_rich_many2many_complete_config_on_self_with_schema(self,
                                                                db_schema):

        def add_in_registry():

            @register(Model)
            class TestLink:
                __db_schema__ = 'test_db_m2m_schema'

                id = Integer(primary_key=True)
                t_left = Integer(foreign_key='Model.Test=>id', nullable=False)
                t_right = Integer(foreign_key='Model.Test=>id', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Test:
                __db_schema__ = 'test_db_m2m_schema'

                id = Integer(primary_key=True)
                childs = Many2Many(
                    model='Model.Test',
                    many2many='parents',
                    join_table="testlink",
                    remote_columns="id", local_columns="id",
                    m2m_local_columns='t_left',
                    m2m_remote_columns='t_right',
                )

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test.insert()
        t1.parents.append(t2)
        link = registry.TestLink.query().one()
        assert link.t_left == t2.id
        assert link.t_right == t1.id
        assert link.id
        assert link.create_at
        assert link.foo == 'bar'

    def test_rich_many2many_complete_config_on_self(self):

        def add_in_registry():

            @register(Model)
            class TestLink:
                id = Integer(primary_key=True)
                t_left = Integer(foreign_key='Model.Test=>id', nullable=False)
                t_right = Integer(foreign_key='Model.Test=>id', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                childs = Many2Many(
                    model='Model.Test',
                    many2many='parents',
                    join_table="testlink",
                    remote_columns="id", local_columns="id",
                    m2m_local_columns='t_left',
                    m2m_remote_columns='t_right',
                )

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test.insert()
        t1.parents.append(t2)
        link = registry.TestLink.query().one()
        assert link.t_left == t2.id
        assert link.t_right == t1.id
        assert link.id
        assert link.create_at
        assert link.foo == 'bar'

    def test_rich_many2many_minimum_config_on_self(self):

        def add_in_registry():

            @register(Model)
            class TestLink:
                id = Integer(primary_key=True)
                t_left = Integer(foreign_key='Model.Test=>id', nullable=False)
                t_right = Integer(foreign_key='Model.Test=>id', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                childs = Many2Many(
                    model='Model.Test',
                    many2many='parents',
                    join_table="testlink",
                    m2m_local_columns='t_left',
                    m2m_remote_columns='t_right',
                )

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test.insert()
        t1.parents.append(t2)
        link = registry.TestLink.query().one()
        assert link.t_left == t2.id
        assert link.t_right == t1.id
        assert link.id
        assert link.create_at
        assert link.foo == 'bar'

    def test_rich_many2many_minimum_config_on_self_with_schema(self, db_schema):

        def add_in_registry():

            @register(Model)
            class TestLink:
                __db_schema__ = 'test_db_m2m_schema'

                id = Integer(primary_key=True)
                t_left = Integer(foreign_key='Model.Test=>id', nullable=False)
                t_right = Integer(foreign_key='Model.Test=>id', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Test:
                __db_schema__ = 'test_db_m2m_schema'

                id = Integer(primary_key=True)
                childs = Many2Many(
                    model='Model.Test',
                    many2many='parents',
                    join_table="testlink",
                    m2m_local_columns='t_left',
                    m2m_remote_columns='t_right',
                )

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test.insert()
        t1.parents.append(t2)
        link = registry.TestLink.query().one()
        assert link.t_left == t2.id
        assert link.t_right == t1.id
        assert link.id
        assert link.create_at
        assert link.foo == 'bar'

    def test_rich_many2many_minimum_config_on_self_without_columns_1(self):

        def add_in_registry():

            @register(Model)
            class TestLink:
                id = Integer(primary_key=True)
                t_left = Integer(foreign_key='Model.Test=>id', nullable=False)
                t_right = Integer(foreign_key='Model.Test=>id', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                childs = Many2Many(
                    model='Model.Test',
                    many2many='parents',
                    join_table="testlink",
                    m2m_local_columns='t_left',
                )

        with pytest.raises(FieldException):
            self.init_registry(add_in_registry)

    def test_rich_many2many_minimum_config_on_self_without_columns_2(self):

        def add_in_registry():

            @register(Model)
            class TestLink:
                id = Integer(primary_key=True)
                t_left = Integer(foreign_key='Model.Test=>id', nullable=False)
                t_right = Integer(foreign_key='Model.Test=>id', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                childs = Many2Many(
                    model='Model.Test',
                    many2many='parents',
                    join_table="testlink",
                    m2m_remote_columns='t_right',
                )

        with pytest.raises(FieldException):
            self.init_registry(add_in_registry)

    def test_rich_many2many_minimum_config_on_self_with_join_model(self):

        def add_in_registry():

            @register(Model)
            class TestLink:
                id = Integer(primary_key=True)
                t_left = Integer(foreign_key='Model.Test=>id', nullable=False)
                t_right = Integer(foreign_key='Model.Test=>id', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                childs = Many2Many(
                    model='Model.Test',
                    many2many='parents',
                    join_model=Model.TestLink,
                    m2m_local_columns='t_left',
                    m2m_remote_columns='t_right',
                )

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test.insert()
        t1.parents.append(t2)
        link = registry.TestLink.query().one()
        assert link.t_left == t2.id
        assert link.t_right == t1.id
        assert link.id
        assert link.create_at
        assert link.foo == 'bar'

    def test_rich_many2many_minimum_config_on_self_with_many2one(self):

        def add_in_registry():

            @register(Model)
            class TestLink:
                id = Integer(primary_key=True)
                left = Many2One(model='Model.Test', nullable=False)
                right = Many2One(model='Model.Test', nullable=False)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                childs = Many2Many(
                    model='Model.Test',
                    many2many='parents',
                    join_model=Model.TestLink,
                    m2m_local_columns='left_id',
                    m2m_remote_columns='right_id',
                )

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test.insert()
        t1.parents.append(t2)
        link = registry.TestLink.query().one()
        assert link.left_id == t2.id
        assert link.right_id == t1.id
        assert link.id
        assert link.create_at
        assert link.foo == 'bar'

    def test_rich_many2many_minimum_config_on_self_with_pk_many2one_2(self):

        def add_in_registry():

            @register(Model)
            class TestLink:
                left = Many2One(
                    model='Model.Test', nullable=False, primary_key=True)
                right = Many2One(
                    model='Model.Test', nullable=False, primary_key=True)
                create_at = DateTime(default=datetime.now)
                foo = String(default='bar')

            @register(Model)
            class Test:

                id = Integer(primary_key=True)
                childs = Many2Many(
                    model='Model.Test',
                    many2many='parents',
                    join_model=Model.TestLink,
                    m2m_local_columns='left',
                    m2m_remote_columns='right',
                )

        registry = self.init_registry(add_in_registry)
        t1 = registry.Test.insert()
        t2 = registry.Test.insert()
        t1.parents.append(t2)
        link = registry.TestLink.query().one()
        assert link.left_id == t2.id
        assert link.right_id == t1.id
        assert link.create_at
        assert link.foo == 'bar'

    def test_with_twice_the_same_many2many(self):
        def add_in_registry(**kwargs):

            @register(Model)
            class Address:

                id = Integer(primary_key=True)
                street = String()
                zip = String()
                city = String()

            @register(Model)
            class Person:

                name = String(primary_key=True)
                invoiced_addresses = Many2Many(model=Model.Address)
                delivery_addresses = Many2Many(model=Model.Address)

        registry = self.init_registry(add_in_registry)
        address = registry.Address.insert(
            street='14-16 rue soleillet', zip='75020', city='Paris')
        person = registry.Person.insert(name="Jean-sébastien SUZANNE")

        person.invoiced_addresses.append(address)

        assert person.invoiced_addresses == [address]
        assert person.delivery_addresses == []

    def test_many2many_and_order_list(self):

        def add_in_registry(**kwargs):

            @register(Model)
            class Address:

                id = Integer(primary_key=True)
                street = String()
                zip = String()
                city = String()

            @register(Model)
            class Person:

                name = String(primary_key=True)
                addresses = Many2Many(
                    model=Model.Address,
                    many2many=(
                        "persons",
                        dict(
                            order_by="ModelPerson.name",
                            collection_class=ordering_list('name'),
                        )
                    ),
                    order_by="ModelAddress.city",
                    collection_class=ordering_list('city'),
                )

        registry = self.init_registry(add_in_registry)

        address3 = registry.Address.insert(city='Paris 3')
        address1 = registry.Address.insert(city='Paris 1')
        address4 = registry.Address.insert(city='Paris 4')
        address2 = registry.Address.insert(city='Paris 2')

        person3 = registry.Person.insert(name="test 3")
        person1 = registry.Person.insert(name="test 1")
        person4 = registry.Person.insert(name="test 4")
        person2 = registry.Person.insert(name="test 2")

        person3.addresses.append(address3)
        person1.addresses.append(address1)
        person4.addresses.append(address4)
        person2.addresses.append(address2)

        person1.addresses.append(address4)
        person1.addresses.append(address2)

        address1.persons.append(person4)

        registry.flush()

        assert address1.persons.ordering_attr == 'name'
        assert person1.addresses.ordering_attr == 'city'

        assert address1.persons == [person1, person4]
        assert address2.persons == [person1, person2]
        assert address3.persons == [person3]
        assert address4.persons == [person1, person4]

        person1.refresh()
        assert person1.addresses == [address1, address2, address4]
        assert person2.addresses == [address2]
        assert person3.addresses == [address3]
        person4.refresh()
        assert person4.addresses == [address1, address4]



# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok project
#
#    Copyright (C) 2014 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2018 Jean-Sebastien SUZANNE <jssuzanne@anybox.fr>
#    Copyright (C) 2021 Jean-Sebastien SUZANNE <js.suzanne@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
import pytest
from .conftest import init_registry
from anyblok.testing import TestCase, LogCapture
from anyblok.registry import (
    RegistryManager, RegistryException, RegistryManagerException,
    Registry)
from anyblok.config import Configuration
from anyblok.blok import BlokManager, Blok
from anyblok.column import Integer
from anyblok.environment import EnvironmentManager
from anyblok import start
from threading import Thread
from logging import ERROR
import sys

try:
    # python 3.4+ should use builtin unittest.mock not mock package
    from unittest.mock import patch
except ImportError:
    from mock import patch


class Test:
    pass


class TestTest:
    pass


class TestTestTest:
    pass


class FakeException(Exception):
    pass


def fake_event(session):
    raise FakeException()


class TestRegistry:

    @pytest.fixture(autouse=True)
    def close_registry(self, request, bloks_loaded):

        def close():
            if hasattr(self, 'registry'):
                self.registry.close()

            for cls in (Test, TestTest):
                if hasattr(cls, 'Test'):
                    delattr(cls, 'Test')

        request.addfinalizer(close)

    def init_registry(self, *args, **kwargs):
        self.registry = init_registry(*args, **kwargs)
        return self.registry

    def test_get_model(self):
        registry = self.init_registry(None)
        System = registry.get('Model.System')
        assert registry.System is System

    def test_get_the_same_registry(self):
        registry = self.init_registry(None)
        registry2 = RegistryManager.get(
            Configuration.get('db_name'),
            unittest=True)
        assert registry is registry2

    def test_reload(self):
        registry = self.init_registry(None)
        bloks_before_reload = registry.System.Blok.query('name').filter(
            registry.System.Blok.state == 'installed').all()
        registry.reload()
        bloks_after_reload = registry.System.Blok.query('name').filter(
            registry.System.Blok.state == 'installed').all()
        assert bloks_before_reload == bloks_after_reload

    def test_get_bloks_to_load(self):
        registry = self.init_registry(None)
        bloks = registry.get_bloks_to_load()
        assert 'anyblok-core' in bloks

    def test_load_entry(self):
        registry = self.init_registry(None)
        registry.loaded_registries['entry_names'] = []
        RegistryManager.loaded_bloks['blok'] = {
            'entry': {
                'registry_names': ['key'],
                'key': {'properties': {'property': True},
                        'bases': [TestCase]},
            },
        }
        registry.load_entry('blok', 'entry')
        assert registry.loaded_registries['key'] == {
            'properties': {'property': True},
            'bases': [TestCase]
        }

    def test_load_core(self):
        RegistryManager.loaded_bloks['blok'] = {
            'Core': {'Session': [TestCase]},
        }
        registry = self.init_registry(None)
        registry.load_core('blok', 'Session')
        assert TestCase in registry.loaded_cores['Session']

    def test_load_blok(self):

        class BlokTest(Blok):
            pass

        registry = self.init_registry(None)
        BlokManager.bloks['blok'] = BlokTest
        RegistryManager.loaded_bloks['blok'] = {
            'Core': {'Session': [TestCase],
                     'Base': [],
                     'SqlBase': [],
                     'SqlViewBase': [],
                     'Query': [],
                     'InstrumentedList': []},
            'removed': [],
            'properties': {},
        }
        for entry in RegistryManager.declared_entries:
            RegistryManager.loaded_bloks['blok'][entry] = {
                'registry_names': []}

        registry.load_blok('blok', False, [])
        assert TestCase in registry.loaded_cores['Session']

    def check_added_in_regisry(self, registry):
        assert registry.Test is Test
        assert registry.Test.Test is TestTest
        assert registry.Test.Test.Test is TestTestTest

    def test_add_in_registry_1(self):
        registry = self.init_registry(None)
        registry.add_in_registry('Declarations.Test', Test)
        registry.add_in_registry('Declarations.Test.Test', TestTest)
        registry.add_in_registry('Declarations.Test.Test.Test', TestTestTest)
        self.check_added_in_regisry(registry)

    def test_start_function(self):
        BlokManager.unload()
        db_name = Configuration.get('db_name') or 'test_anyblok'
        db_driver_name = Configuration.get('db_driver_name') or 'postgresql'

        testargs = ['default', '--db-name', db_name, '--db-driver-name',
                    db_driver_name]
        with patch.object(sys, 'argv', testargs):
            registry = start('default')

        assert registry is not None

    def test_add_in_registry_2(self):
        registry = self.init_registry(None)
        registry.add_in_registry('Declarations.Test', Test)
        registry.add_in_registry('Declarations.Test.Test.Test', TestTestTest)
        registry.add_in_registry('Declarations.Test.Test', TestTest)
        self.check_added_in_regisry(registry)

    def test_add_in_registry_3(self):
        registry = self.init_registry(None)
        registry.add_in_registry('Declarations.Test.Test', TestTest)
        registry.add_in_registry('Declarations.Test', Test)
        registry.add_in_registry('Declarations.Test.Test.Test', TestTestTest)
        self.check_added_in_regisry(registry)

    def test_add_in_registry_4(self):
        registry = self.init_registry(None)
        registry.add_in_registry('Declarations.Test.Test', TestTest)
        registry.add_in_registry('Declarations.Test.Test.Test', TestTestTest)
        registry.add_in_registry('Declarations.Test', Test)
        self.check_added_in_regisry(registry)

    def test_add_in_registry_5(self):
        registry = self.init_registry(None)
        registry.add_in_registry('Declarations.Test.Test.Test', TestTestTest)
        registry.add_in_registry('Declarations.Test', Test)
        registry.add_in_registry('Declarations.Test.Test', TestTest)
        self.check_added_in_regisry(registry)

    def test_add_in_registry_6(self):
        registry = self.init_registry(None)
        registry.add_in_registry('Declarations.Test.Test.Test', TestTestTest)
        registry.add_in_registry('Declarations.Test.Test', TestTest)
        registry.add_in_registry('Declarations.Test', Test)
        self.check_added_in_regisry(registry)

    def test_registry_db_exist(self):
        assert Configuration.get('Registry').db_exists(
            db_name=Configuration.get('db_name'))

    def test_registry_db_unexist(self):
        assert not (Configuration.get('Registry').db_exists(
            db_name='wrong_db_name'))


class TestRegistry2:

    @pytest.fixture(autouse=True)
    def close_registry(self, request, bloks_loaded):

        def close():
            if hasattr(self, 'registry'):
                self.registry.close()

            for cls in (Test, TestTest):
                if hasattr(cls, 'Test'):
                    delattr(cls, 'Test')

        request.addfinalizer(close)

    def init_registry(self, *args, **kwargs):
        self.registry = init_registry(*args, **kwargs)
        return self.registry

    def add_model(self):

        from anyblok import Declarations

        register = Declarations.register
        Model = Declarations.Model

        @register(Model)
        class Test:

            id = Integer(primary_key=True)
            val = Integer(default=0)

            @classmethod
            def _precommit_hook(cls):
                for t in cls.query().all():
                    t.val += t.id

            def add_precommit_hook(self):
                self.precommit_hook('_precommit_hook')

            @classmethod
            def add_cl_precommit_hook(cls):
                cls.precommit_hook('_precommit_hook')

    def test_precommit_hook(self):
        registry = self.init_registry(self.add_model)
        t1 = registry.Test.insert()
        t1.add_precommit_hook()
        t2 = registry.Test.insert()
        t2.add_precommit_hook()
        assert t1.val == 0
        assert t2.val == 0
        registry.commit()
        assert t1.val == t1.id
        assert t2.val == t2.id
        registry.commit()
        assert t1.val == t1.id
        assert t2.val == t2.id
        registry.Test.add_cl_precommit_hook()
        assert t1.val == t1.id
        assert t2.val == t2.id
        registry.commit()
        assert t1.val == 2 * t1.id
        assert t2.val == 2 * t2.id
        registry.commit()
        assert t1.val == 2 * t1.id
        assert t2.val == 2 * t2.id
        t1.add_precommit_hook()
        assert t1.val == 2 * t1.id
        assert t2.val == 2 * t2.id
        registry.commit()
        assert t1.val == 3 * t1.id
        assert t2.val == 3 * t2.id

    def test_precommit_hook_in_thread(self):
        registry = self.init_registry(self.add_model)
        t1 = registry.Test.insert()
        t1.add_precommit_hook()
        t2 = registry.Test.insert()
        t2.add_precommit_hook()
        assert t1.val == 0
        assert t2.val == 0

        def target():
            registry.commit()

        t = Thread(target=target)
        t.start()
        t.join()

        assert t1.val == 0
        assert t2.val == 0
        registry.commit()
        assert t1.val == t1.id
        assert t2.val == t2.id

    def test_precommit_hook_in_thread_2(self):
        do_somthing = 0

        def add_in_registry():

            from anyblok import Declarations

            @Declarations.register(Declarations.Model)
            class Test:

                @classmethod
                def _precommit_hook(cls):
                    nonlocal do_somthing
                    do_somthing += 1

        registry = self.init_registry(add_in_registry)
        assert do_somthing == 0

        def target():
            registry.Test.precommit_hook('_precommit_hook')
            registry.commit()

        t = Thread(target=target)
        t.start()
        t.join()

        registry.commit()
        assert do_somthing == 1

    def define_cls(self, typename='Model', name='Test', val=1, usesuper=False,
                   inherit=None):

        from anyblok import Declarations

        register = Declarations.register
        Type = getattr(Declarations, typename)
        if inherit is None:
            inherit = object
        else:
            inherit = getattr(Declarations.Mixin, inherit)

        @register(Type, name_=name)
        class Test(inherit):

            @classmethod
            def foo(cls):
                if usesuper:
                    return val * super(Test, cls).foo()

                return val

        return Test

    def test_check_define_cls(self):

        def add_in_registry():
            self.define_cls()

        registry = self.init_registry(add_in_registry)
        assert registry.Test.foo() == 1

    def test_check_define_cls_with_inherit(self):

        def add_in_registry():
            self.define_cls()
            self.define_cls(val=2, usesuper=True)

        registry = self.init_registry(add_in_registry)
        assert registry.Test.foo() == 2

    def test_check_define_cls_with_inherit_core(self):

        def add_in_registry():
            self.define_cls(typename='Core', name='Base', val=2)
            self.define_cls(val=2, usesuper=True)

        registry = self.init_registry(add_in_registry)
        assert registry.Test.foo() == 4

    def test_check_define_cls_with_inherit_mixin(self):

        def add_in_registry():
            self.define_cls(typename='Mixin', name='MTest', val=3)
            self.define_cls(val=3, usesuper=True, inherit='MTest')

        registry = self.init_registry(add_in_registry)
        assert registry.Test.foo() == 9

    def test_check_define_cls_with_inherit2(self):

        def add_in_registry():
            self.define_cls()
            self.define_cls(val=2, usesuper=True)
            self.define_cls(val=2, usesuper=True)

        registry = self.init_registry(add_in_registry)
        assert registry.Test.foo() == 4

    def test_check_define_cls_with_inherit_core2(self):

        def add_in_registry():
            self.define_cls(typename='Core', name='Base', val=2)
            self.define_cls(typename='Core', name='Base', val=2, usesuper=True)
            self.define_cls(val=2, usesuper=True)

        registry = self.init_registry(add_in_registry)
        assert registry.Test.foo() == 8

    def test_check_define_cls_with_inherit_mixin2(self):

        def add_in_registry():
            self.define_cls(typename='Mixin', name='MTest', val=3)
            self.define_cls(typename='Mixin', name='MTest', val=3,
                            usesuper=True)
            self.define_cls(val=3, usesuper=True, inherit='MTest')

        registry = self.init_registry(add_in_registry)
        assert registry.Test.foo() == 27

    def test_remove(self):

        def add_in_registry():
            self.define_cls()
            cls_ = self.define_cls(val=2, usesuper=True)
            self.define_cls(val=2, usesuper=True)
            from anyblok import Declarations
            Declarations.unregister(Declarations.Model.Test, cls_)

        registry = self.init_registry(add_in_registry)
        assert registry.Test.foo() == 2

    def test_remove_core(self):

        def add_in_registry():
            self.define_cls(typename='Core', name='Base', val=2)
            cls_ = self.define_cls(typename='Core', name='Base', val=2,
                                   usesuper=True)
            self.define_cls(val=2, usesuper=True)
            from anyblok import Declarations
            Declarations.unregister(Declarations.Core.Base, cls_)

        registry = self.init_registry(add_in_registry)
        assert registry.Test.foo() == 4

    def test_remove_mixin(self):

        def add_in_registry():
            self.define_cls(typename='Mixin', name='MTest', val=3)
            cls_ = self.define_cls(typename='Mixin', name='MTest', val=3,
                                   usesuper=True)
            self.define_cls(val=3, usesuper=True, inherit='MTest')
            from anyblok import Declarations
            Declarations.unregister(Declarations.Mixin.MTest, cls_)

        registry = self.init_registry(add_in_registry)
        assert registry.Test.foo() == 9

    def test_postcommit_hook(self):
        do_somthing = 0

        def add_in_registry():

            from anyblok import Declarations

            @Declarations.register(Declarations.Model)
            class Test:

                id = Integer(primary_key=True)

                @classmethod
                def _postcommit_hook(cls):
                    nonlocal do_somthing
                    do_somthing += 1

                def add_precommit_hook(self):
                    self.postcommit_hook('_postcommit_hook')

                @classmethod
                def add_cl_precommit_hook(cls):
                    cls.postcommit_hook('_postcommit_hook')

        registry = self.init_registry(add_in_registry)
        t = registry.Test.insert()
        t.add_precommit_hook()
        assert do_somthing == 0
        registry.commit()
        assert do_somthing == 1
        registry.commit()
        assert do_somthing == 1
        registry.Test.add_cl_precommit_hook()
        assert do_somthing == 1
        registry.commit()
        assert do_somthing == 2
        registry.commit()
        assert do_somthing == 2
        t.add_precommit_hook()
        assert do_somthing == 2
        registry.commit()
        assert do_somthing == 3

    def test_postcommit_hook_twice(self):

        def add_in_registry():

            from anyblok import Declarations

            @Declarations.register(Declarations.Model)
            class Test:
                pass

        registry = self.init_registry(add_in_registry)
        registry.Test.postcommit_hook('_postcommit_hook')
        registry.Test.postcommit_hook('_postcommit_hook')
        assert len(EnvironmentManager.get('_postcommit_hook', [])) == 1

    def test_postcommit_hook_twice_with_after_another(self):

        def add_in_registry():

            from anyblok import Declarations

            @Declarations.register(Declarations.Model)
            class Test:
                pass

        registry = self.init_registry(add_in_registry)
        registry.Test.postcommit_hook('_postcommit_hook1')
        registry.Test.postcommit_hook('_postcommit_hook2')
        assert len(EnvironmentManager.get('_postcommit_hook', [])) == 2
        assert [x[1] for x in EnvironmentManager.get('_postcommit_hook')] == [
            '_postcommit_hook1', '_postcommit_hook2']
        registry.Test.postcommit_hook('_postcommit_hook1')
        assert [x[1] for x in EnvironmentManager.get('_postcommit_hook')] == [
            '_postcommit_hook1', '_postcommit_hook2']
        registry.Test.postcommit_hook(
            '_postcommit_hook1', put_at_the_end_if_exist=True)
        assert [x[1] for x in EnvironmentManager.get('_postcommit_hook')] == [
            '_postcommit_hook2', '_postcommit_hook1']

    def test_postcommit_hook_in_thread(self):
        do_somthing = 0

        def add_in_registry():

            from anyblok import Declarations

            @Declarations.register(Declarations.Model)
            class Test:

                @classmethod
                def _postcommit_hook(cls):
                    nonlocal do_somthing
                    do_somthing += 1

        registry = self.init_registry(add_in_registry)
        registry.Test.postcommit_hook('_postcommit_hook')
        assert do_somthing == 0

        def target():
            registry.commit()

        t = Thread(target=target)
        t.start()
        t.join()

        assert do_somthing == 0
        registry.commit()
        assert do_somthing == 1

    def test_postcommit_hook_in_thread_2(self):
        do_somthing = 0

        def add_in_registry():

            from anyblok import Declarations

            @Declarations.register(Declarations.Model)
            class Test:

                @classmethod
                def _postcommit_hook(cls):
                    nonlocal do_somthing
                    do_somthing += 1

        registry = self.init_registry(add_in_registry)
        assert do_somthing == 0

        def target():
            registry.Test.postcommit_hook('_postcommit_hook')
            registry.commit()

        t = Thread(target=target)
        t.start()
        t.join()

        assert do_somthing == 1

    def test_postcommit_hook_call_only_if_commited(self):
        do_somthing = 0

        def add_in_registry():

            from anyblok import Declarations

            @Declarations.register(Declarations.Model)
            class Test:

                @classmethod
                def _precommit_hook(cls):
                    raise Exception('Test')

                @classmethod
                def _postcommit_hook(cls):
                    nonlocal do_somthing
                    do_somthing += 1

        registry = self.init_registry(add_in_registry)
        registry.Test.postcommit_hook('_postcommit_hook')
        assert do_somthing == 0
        registry.commit()
        assert do_somthing == 1
        registry.Test.precommit_hook('_precommit_hook')
        registry.Test.postcommit_hook('_postcommit_hook')
        with pytest.raises(Exception):
            registry.commit()

        assert do_somthing == 1

    def test_postcommit_hook_call_only_if_raised(self):
        do_somthing = 0

        def add_in_registry():

            from anyblok import Declarations

            @Declarations.register(Declarations.Model)
            class Test:

                @classmethod
                def _precommit_hook(cls):
                    raise Exception('Test')

                @classmethod
                def _postcommit_hook(cls):
                    nonlocal do_somthing
                    do_somthing += 1

        registry = self.init_registry(add_in_registry)
        registry.Test.postcommit_hook(
            '_postcommit_hook', call_only_if='raised')
        assert do_somthing == 0
        registry.commit()
        assert do_somthing == 0
        registry.Test.precommit_hook('_precommit_hook')
        registry.Test.postcommit_hook(
            '_postcommit_hook', call_only_if='raised')
        with pytest.raises(Exception):
            registry.commit()

        assert do_somthing == 1

    def test_postcommit_hook_call_only_if_always(self):
        do_somthing = 0

        def add_in_registry():

            from anyblok import Declarations

            @Declarations.register(Declarations.Model)
            class Test:

                @classmethod
                def _precommit_hook(cls):
                    raise Exception('Test')

                @classmethod
                def _postcommit_hook(cls):
                    nonlocal do_somthing
                    do_somthing += 1

        registry = self.init_registry(add_in_registry)
        registry.Test.postcommit_hook(
            '_postcommit_hook', call_only_if='always')
        assert do_somthing == 0
        registry.commit()
        assert do_somthing == 1
        registry.Test.precommit_hook('_precommit_hook')
        registry.Test.postcommit_hook(
            '_postcommit_hook', call_only_if='always')
        with pytest.raises(Exception):
            registry.commit()

        assert do_somthing == 2

    def test_postcommit_hook_with_rollback(self):
        do_somthing = 0

        def add_in_registry():

            from anyblok import Declarations

            @Declarations.register(Declarations.Model)
            class Test:

                @classmethod
                def _postcommit_hook(cls):
                    nonlocal do_somthing
                    do_somthing += 1

        registry = self.init_registry(add_in_registry)
        registry.Test.postcommit_hook(
            '_postcommit_hook')
        assert do_somthing == 0
        registry.rollback()
        registry.commit()
        assert do_somthing == 0

    def test_postcommit_hook_with_exception(self):

        def add_in_registry():

            from anyblok import Declarations

            @Declarations.register(Declarations.Model)
            class Test:

                @classmethod
                def _postcommit_hook(cls):
                    raise Exception('Here one exception')

        registry = self.init_registry(add_in_registry)
        registry.Test.postcommit_hook(
            '_postcommit_hook')
        with LogCapture('anyblok.registry', level=ERROR) as logs:
            registry.commit()
            messages = logs.get_error_messages()
            message = messages[0]
            assert 'Here one exception' in message

    def test_precommit_hook_twice(self):

        def add_in_registry():

            from anyblok import Declarations

            @Declarations.register(Declarations.Model)
            class Test:
                pass

        registry = self.init_registry(add_in_registry)
        registry.Test.precommit_hook('_precommit_hook')
        registry.Test.precommit_hook('_precommit_hook')
        assert len(EnvironmentManager.get('_precommit_hook', [])) == 1

    def test_precommit_hook_twice_with_after_another(self):

        def add_in_registry():

            from anyblok import Declarations

            @Declarations.register(Declarations.Model)
            class Test:
                pass

        registry = self.init_registry(add_in_registry)
        registry.Test.precommit_hook('_precommit_hook1')
        registry.Test.precommit_hook('_precommit_hook2')
        assert len(EnvironmentManager.get('_precommit_hook', [])) == 2
        assert [x[1] for x in EnvironmentManager.get('_precommit_hook')] == [
            '_precommit_hook1', '_precommit_hook2']
        registry.Test.precommit_hook('_precommit_hook1')
        assert [x[1] for x in EnvironmentManager.get('_precommit_hook')] == [
            '_precommit_hook1', '_precommit_hook2']
        registry.Test.precommit_hook(
            '_precommit_hook1', put_at_the_end_if_exist=True)
        assert [x[1] for x in EnvironmentManager.get('_precommit_hook')] == [
            '_precommit_hook2', '_precommit_hook1']


class TestRegistry3:

    def test_refresh(self, registry_blok):
        registry = registry_blok
        blok = registry.System.Blok.query().get('anyblok-core')
        registry.refresh(blok)

    def test_refresh_attribute_name(self, registry_blok):
        registry = registry_blok
        blok = registry.System.Blok.query().get('anyblok-core')
        registry.refresh(blok, attribute_names=['name', 'version'])

    def test_expurge(self, registry_blok):
        registry = registry_blok
        blok = registry.System.Blok.query().get('anyblok-core')
        registry.expunge(blok)

    def test_apply_session_events_from_setting(self, registry_blok):
        registry = registry_blok

        additional_setting = registry.additional_setting.get(
            'anyblok.session.event', [])
        additional_setting.append(fake_event)
        registry.additional_setting[
            'anyblok.session.event'] = additional_setting
        with pytest.raises(FakeException):
            registry.apply_session_events()

        additional_setting.remove(fake_event)

    def test_get(self, registry_blok):
        registry = registry_blok
        assert registry.get(
            'Model.System.Blok'
        ) is registry.System.Blok

    def test_get_unexisting_model(self, registry_blok):
        registry = registry_blok
        with pytest.raises(RegistryManagerException):
            registry.get('Model.Unexisting.Model')

    def test_get_bloks_by_states_without_state(self, registry_blok):
        registry = registry_blok
        assert registry.get_bloks_by_states() == []

    def test_db_exists_without_db_name(self):
        with pytest.raises(RegistryException):
            Registry.db_exists()

    def test_apply_engine_events_from_setting(self, registry_blok):
        registry = registry_blok

        additional_setting = registry.additional_setting.get(
            'anyblok.engine.event', [])
        additional_setting.append(fake_event)
        registry.additional_setting[
            'anyblok.engine.event'] = additional_setting
        with pytest.raises(FakeException):
            registry.apply_engine_events(registry.rw_engine)

        additional_setting.remove(fake_event)

    def test_apply_state_on_an_unexisting_blok(self, registry_blok):
        registry = registry_blok
        with pytest.raises(RegistryException):
            registry.apply_state(
                'unexisting_blok', 'toinstall', ['toinstall'])

    def test_apply_the_same_state(self, registry_blok):
        registry = registry_blok
        assert registry.apply_state(
            'anyblok-core', 'installed', ['installed']) is None



# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok project
#
#    Copyright (C) 2019 Joachim Trouverie
#    Copyright (C) 2020 Jean-Sebastien SUZANNE <js.suzanne@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from sqlalchemy.dialects.mysql.types import TINYINT
from sqlalchemy.dialects.mssql.base import BIT
from sqlalchemy.sql.sqltypes import Boolean, DateTime
from logging import getLogger
from .migration import MigrationColumnTypePlugin
logger = getLogger(__name__)


class BooleanToTinyIntMySQL(MigrationColumnTypePlugin):

    to_type = Boolean
    from_type = TINYINT
    dialect = ['MySQL', 'MariaDB']

    def need_to_modify_type(self):
        '''Boolean are TINYINT in MySQL DataBases'''
        return False

    def apply(self, column, **kwargs):
        '''Boolean are TINYINT in MySQL DataBases'''
        # do nothing
        pass  # pragma: no cover


class DateTimeToDateTimeMySQL(MigrationColumnTypePlugin):

    to_type = DateTime
    from_type = DateTime
    dialect = ['MySQL', 'MariaDB']

    def need_to_modify_type(self):
        return False

    def apply(self, column, **kwargs):
        pass  # pragma: no cover


class BooleanToBitMsSQL(MigrationColumnTypePlugin):

    to_type = Boolean
    from_type = BIT
    dialect = ['MsSQL']

    def need_to_modify_type(self):
        '''Boolean are Bit in MsSQL DataBases'''
        return False

    def apply(self, column, **kwargs):
        '''Boolean are Bit in MsSQL DataBases'''
        # do nothing
        pass  # pragma: no cover



# -*- coding: utf-8 -*-
# This file is a part of the AnyBlok project
#
#    Copyright (C) 2021 Jean-Sebastien SUZANNE <js.suzanne@gmail.com>
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file,You can
# obtain one at http://mozilla.org/MPL/2.0/.
from sqlalchemy import event


def mysql_no_autocommit(engine):

    def mysql_set_no_autocommit(dbapi_con, connection_record):
        cur = dbapi_con.cursor()
        cur.execute("SET autocommit=0;")
        cur.execute("SET SESSION sql_mode='TRADITIONAL';")
        cur = None

    event.listen(engine, 'connect', mysql_set_no_autocommit)
